/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESTrees.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */
package weka.classifiers.meta;

import weka.core.*;
import java.util.*;
import weka.classifiers.meta.baggingEnsembleSelection.*;
import weka.classifiers.trees.REPTree;
import weka.classifiers.Classifier;
import weka.classifiers.AbstractClassifier;
import weka.core.AdditionalMeasureProducer;
import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Randomizable;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import java.util.Enumeration;
import weka.classifiers.meta.baggingEnsembleSelection.BESHelper.BESBaseClassifier;
import weka.classifiers.meta.baggingEnsembleSelection.BESHelper.BESBaseClassifierLib;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 */
public class BESTrees
        extends AbstractClassifier
        implements OptionHandler, Randomizable, WeightedInstancesHandler,
        AdditionalMeasureProducer, TechnicalInformationHandler {

    private BESoob m_BES = new BESoob();
    protected boolean m_UseNNLS = false;
    protected int m_NumBags = 10;
    protected int m_NumTreesPerBag = 10;
    protected int m_NumFeatures = 0;
    protected int m_RandomSeed = 1;
    protected int m_KValue = 0;
    protected int m_FinalEnsembleSize = -1;
    protected int m_Slots = 1;
    protected int m_MaxDepth = -1;
    protected int m_NumFolds = 3;
    protected int m_HillclimbMetric = BESHelper.METRIC_RMSE;

    public SelectedTag getHillclimbMetricMethod() {
        return new SelectedTag(m_HillclimbMetric, BESDirectHillclimbingES.TAGS_METRIC);
    }

    public void setHillclimbMetricMethod(SelectedTag method) {
        if (method.getTags() == BESDirectHillclimbingES.TAGS_METRIC) {
            m_HillclimbMetric = method.getSelectedTag().getID();
        }
    }

    public void setNumExecutionSlots(int numSlots) {
        m_Slots = numSlots;
    }

    public int getNumExecutionSlots() {
        return m_Slots;
    }

    public int getNumFeatures() {
        return m_NumFeatures;
    }

    public void setNumFeatures(int newNumFeatures) {
        m_NumFeatures = newNumFeatures;
    }

    public int getMaxDepth() {
        return m_MaxDepth;
    }

    public void setMaxDepth(int value) {
        m_MaxDepth = value;
        if (m_MaxDepth <= 0) {
            m_MaxDepth = -1;
        }
    }

    @Override
    public Capabilities getCapabilities() {
        return new REPTree().getCapabilities();
    }

    @Override
    public void setSeed(int seed) {
        m_RandomSeed = seed;
    }

    @Override
    public int getSeed() {
        return m_RandomSeed;
    }

    public int getFinalEnsembleSize() {
        if (m_UseNNLS) {
            return m_FinalEnsembleSize;
        } else {
            return m_NumBags;
        }
    }

    public void setNumTreesPerBag(int n) {
        if (n < 2) {
            m_NumTreesPerBag = 2;
        } else {
            m_NumTreesPerBag = n;
        }
    }

    public int getNumTreesPerBag() {
        return m_NumTreesPerBag;
    }

    public int getTotalNumberTrees() {
        return m_BES.getTotalNumberTrees();
    }

    public void setNumBags(int n) {
        if (n < 1) {
            m_NumBags = 1;
        } else {
            m_NumBags = n;
        }
    }

    public int getNumBags() {
        return m_NumBags;
    }

    public void setUseNNLS(boolean use) {
        m_UseNNLS = use;
    }

    public boolean getUseNNLS() {
        return m_UseNNLS;
    }

    public void setOptions(String[] options) throws Exception {

        String selectionString = Utils.getOption('Z', options);
        if (selectionString.length() != 0) {
            setHillclimbMetricMethod(new SelectedTag(Integer
                    .parseInt(selectionString),
                    BESDirectHillclimbingES.TAGS_METRIC));
        } else {
            setHillclimbMetricMethod(new SelectedTag(BESHelper.METRIC_RMSE,
                    BESDirectHillclimbingES.TAGS_METRIC));
        }

        String tmpStr;

        tmpStr = Utils.getOption('I', options);
        if (tmpStr.length() != 0) {
            m_NumBags = Integer.parseInt(tmpStr);
        } else {
            m_NumBags = 10;
        }

        tmpStr = Utils.getOption('H', options);
        if (tmpStr.length() != 0) {
            m_NumTreesPerBag = Integer.parseInt(tmpStr);
        } else {
            m_NumTreesPerBag = 10;
        }

        tmpStr = Utils.getOption('K', options);
        if (tmpStr.length() != 0) {
            m_NumFeatures = Integer.parseInt(tmpStr);
        } else {
            m_NumFeatures = 0;
        }

        tmpStr = Utils.getOption('G', options);
        if (tmpStr.length() != 0) {
            m_Slots = Integer.parseInt(tmpStr);
        } else {
            m_Slots = 1;
        }

        tmpStr = Utils.getOption('S', options);
        if (tmpStr.length() != 0) {
            setSeed(Integer.parseInt(tmpStr));
        } else {
            setSeed(1);
        }

        tmpStr = Utils.getOption("depth", options);
        if (tmpStr.length() != 0) {
            setMaxDepth(Integer.parseInt(tmpStr));
        } else {
            setMaxDepth(0);
        }

        setUseNNLS(Utils.getFlag('N', options));

        super.setOptions(options);

        Utils.checkForRemainingOptions(options);
    }

    public String[] getOptions() {
        Vector result;
        String[] options;
        int i;

        result = new Vector();

        result.add("-Z");
        result.add("" + getHillclimbMetricMethod().getSelectedTag().getID());

        result.add("-I");
        result.add("" + getNumBags());

        result.add("-H");
        result.add("" + getNumTreesPerBag());

        result.add("-K");
        result.add("" + getNumFeatures());

        result.add("-S");
        result.add("" + getSeed());

        result.add("-G");
        result.add("" + this.getNumExecutionSlots());

        if (getUseNNLS()) {
            result.add("-N");
        }

        if (getMaxDepth() > 0) {
            result.add("-depth");
            result.add("" + getMaxDepth());
        }

        options = super.getOptions();
        for (i = 0; i < options.length; i++) {
            result.add(options[i]);
        }

        return (String[]) result.toArray(new String[result.size()]);
    }

    public void setNumFolds(int n) {
        m_NumFolds = n;
    }

    public void buildClassifier(Instances data) throws Exception {
        // can classifier handle the data?
        getCapabilities().testWithFail(data);

        if (data.attribute(data.classIndex()).isNominal()) {
            m_UseNNLS = false;
        }

        // remove instances with missing class
        data = new Instances(data);
        data.deleteWithMissingClass();

        Random r = new Random(this.m_RandomSeed);

        // set up the random tree options
        m_KValue = m_NumFeatures;
        if (m_KValue < 1) {
            m_KValue = (int) Utils.log2(data.numAttributes()) + 1;
        }
        double attRatio = 1.0 * m_KValue / (data.numAttributes() - 1);

        // construct a base classifier lib
        BESHelper.BESBaseClassifierLib lib = new BESHelper.BESBaseClassifierLib();
        lib.setDebug(false);

        ArrayList<String> algos = new ArrayList<String>();
        for (int i = 1; i <= m_NumTreesPerBag; i++) {
            algos.add("weka.classifiers.meta.RandomSubSpace -P "
                    + attRatio + " -S " + r.nextInt()
                    + " -num-slots 1 -I 1 -W weka.classifiers.trees.REPTree -- -M 2 -V 0.0010 -N "
                    + m_NumFolds + " -S " + r.nextInt() + " -L " + m_MaxDepth);
        }

        for (int run = 0; run < algos.size(); run++) {

            String algoStr = algos.get(run);
            String[] parts = algoStr.split(" ");
            int firstSpace = algoStr.indexOf(" ");
            String opStr = algoStr.substring(firstSpace + 1, algoStr.length());
            try {
                String[] ops = weka.core.Utils.splitOptions(opStr);
                Classifier model = AbstractClassifier.forName(parts[0], ops);
                String threadID = "algo" + run;
                BESHelper.BESBaseClassifier bcModel = new BESHelper.BESBaseClassifier(threadID, model);
                lib.add(bcModel);
            } catch (Exception e) {
                System.err.println(algoStr + ": " + e.toString());
            }
        }

        // Setup DHES
        BESDirectHillclimbingES des = new BESDirectHillclimbingES();
        des.setBaseClassifierLibrary(lib);
        des.setDebug(false);

        des.setInternalHillclimbSetRatio(0.33);
        des.setHillclimbMetric(getHillclimbMetricMethod());
        des.setHillclimbIterations(lib.size());

        BESoob besOOB = new BESoob();

        besOOB.setNumExecutionSlots(m_Slots);
        besOOB.setNumIterations(m_NumBags);

        besOOB.setClassifier(des);
        besOOB.setCalcOutOfBag(true);
        besOOB.setDebug(m_Debug);
        besOOB.setUseNNLS(m_UseNNLS);

        besOOB.buildClassifier(data);

        m_BES = besOOB;
        m_FinalEnsembleSize = m_BES.getFinalEnsembleSize();
    }

    public double[] distributionForInstance(Instance instance) throws Exception {
        return m_BES.distributionForInstance(instance);
    }

    public TechnicalInformation getTechnicalInformation() {
        TechnicalInformation result;
        result = new TechnicalInformation(Type.INPROCEEDINGS);
        result.setValue(Field.AUTHOR, "Quan Sun and Bernhard Pfahringer");
        result.setValue(Field.YEAR, "2012");
        result.setValue(Field.TITLE, "Bagging Ensemble Selection for Regression");
        result.setValue(Field.JOURNAL, "In Proceedings of the 25th Australasian Joint Conference on Artificial Intelligence (AI'12)");
        result.setValue(Field.PAGES, "695-706");
        return result;
    }

    public Enumeration enumerateMeasures() {

        Vector newVector = new Vector(2);
        newVector.addElement("measureFinalEnsembleSize");
        newVector.addElement("measureFinalNumTrees");
        return newVector.elements();
    }

    public double getMeasure(String additionalMeasureName) {

        if (additionalMeasureName.equalsIgnoreCase("measureFinalEnsembleSize")) {
            return measureFinalEnsembleSize();
        } else if (additionalMeasureName.equalsIgnoreCase("measureFinalNumTrees")) {
            return measureFinalNumTrees();
        } else {
            throw new IllegalArgumentException(additionalMeasureName + " not supported (RandomForest)");
        }
    }

    public double measureFinalEnsembleSize() {
        return getFinalEnsembleSize();
    }

    public double measureFinalNumTrees() {
        return m_BES.getTotalNumberTrees();
    }

    public Enumeration listOptions() {

        Vector newVector = new Vector();

        newVector.addElement(new Option(
                "\tNumber of trees to build.",
                "I", 1, "-I <number of trees>"));

        newVector.addElement(new Option(
                "\tNumber of features to consider (<1=int(logM+1)).",
                "K", 1, "-K <number of features>"));

        newVector.addElement(new Option(
                "\tSeed for random number generator.\n" + "\t(default 1)",
                "S", 1, "-S"));

        newVector.addElement(new Option(
                "\tThe maximum depth of the trees, 0 for unlimited.\n" + "\t(default 0)",
                "depth", 1, "-depth <num>"));
        newVector.addElement(
                new Option(
                "\tSet the target metric"
                + " to use. 0 = Correlation Coefficient, 1 = RMSE, 2 = ROC, "
                + "3 = PRECISION, 4 = Recall, 5 = F score, 6 = All, "
                + "7 = Weighted TP, 8 = Mean Abs Error, 9 = Accuracy \n"
                + "\t(default 1 = RMSE)",
                "Z", 2, "-Z <target metric>"));


        Enumeration enu = super.listOptions();
        while (enu.hasMoreElements()) {
            newVector.addElement(enu.nextElement());
        }

        return newVector.elements();
    }

    public String toString() {

        if (m_BES == null) {
            return "BESTrees: No model built yet.";
        }
        int totalNumTrees = getNumTreesPerBag() * getNumBags();
        String result = "The final BESTrees model has " + this.getTotalNumberTrees() + " trees (out of " + totalNumTrees + ") \n";
        return result;
    }

    /**
     * Main method for testing this class.
     *
     * @param argv the options
     */
    public static void main(String[] argv) throws Exception {
        runClassifier(new BESTrees(), argv);
    }
}
