/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESDirectHillclimbingES.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta.baggingEnsembleSelection;

import java.util.*;
import weka.classifiers.*;
import weka.classifiers.meta.baggingEnsembleSelection.BESHelper.BESBaseClassifier;
import weka.classifiers.meta.baggingEnsembleSelection.BESHelper.BESBaseClassifierLib;
import weka.core.*;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 */
public class BESDirectHillclimbingES extends RandomizableClassifier
        implements AdditionalMeasureProducer, TechnicalInformationHandler {

    public static final Tag[] TAGS_METRIC = {
        new Tag(BESHelper.METRIC_ACCURACY, "Optimize Accuracy"),
        new Tag(BESHelper.METRIC_CC, "Optimize Correlation Coefficient"),
        new Tag(BESHelper.METRIC_RMSE, "Optimize with RMSE"),
        new Tag(BESHelper.METRIC_ROC, "Optimize with ROC"),
        new Tag(BESHelper.METRIC_PRECISION, "Optimize with precision"),
        new Tag(BESHelper.METRIC_RECALL, "Optimize with recall"),
        new Tag(BESHelper.METRIC_FSCORE, "Optimize with fscore"),
        new Tag(BESHelper.METRIC_ALL, "Optimize with all metrics"),
        new Tag(BESHelper.METRIC_WEIGHTEDTP, "Optimize with weighted TP")};
    
    protected Instances m_ExternalHillclimbSet = null;
    protected double m_RatioInternalHillclimbSet = 0.1;
    protected boolean m_RandomizeDatset = false;    
    protected BESBaseClassifierLib m_BCLibrary = null;
    protected int m_hillclimbMetric = BESHelper.METRIC_MAE;
    protected int m_hillclimbIterations = 2;
    protected BESBaseClassifier[] m_chosen_models = null;
    protected int[] m_chosen_model_weights = null;
    protected int m_total_weight = 0;
    protected int m_FinalEnsembleSize = 0;
    protected boolean m_UseBestModel = false;
    protected int m_numModelBags = 1;
    protected double m_bestPerformance = 0;

    public int getNumModelBags() {
        return m_numModelBags;
    }

    public void setNumModelBags(int n) throws Exception {
        if (n <= 0) {
            throw new IllegalArgumentException(
                    "EnsembleSelection: Number of model bags "
                    + "must be positive.");
        }
        m_numModelBags = n;
    }

    public void setUseBestModel(boolean use) {
        m_UseBestModel = use;
    }

    public Enumeration enumerateMeasures() {
        Vector newVector = new Vector(2);
        newVector.addElement("measureFinalEnsembleSize");
        return newVector.elements();
    }

    public double getMeasure(String additionalMeasureName) {
        if (additionalMeasureName.equalsIgnoreCase("measureFinalEnsembleSize")) {
            return getFinalEnsembleSize();
        } else {
            throw new IllegalArgumentException(additionalMeasureName + " not supported (Bagging)");
        }
    }

    public int getFinalEnsembleSize() {
        return m_FinalEnsembleSize;
    }

    public double getFinalEnsembleSelectionPercentage() {
        return 1.0 * m_FinalEnsembleSize / m_BCLibrary.size();
    }

    public TechnicalInformation getTechnicalInformation() {
        TechnicalInformation result;
        result = new TechnicalInformation(Type.INPROCEEDINGS);
        result.setValue(Field.AUTHOR, "Rich Caruana, Alex Niculescu, Geoff Crew, and Alex Ksikes");
        result.setValue(Field.TITLE, "Ensemble Selection from Libraries of Models");
        result.setValue(Field.BOOKTITLE, "21st International Conference on Machine Learning");
        result.setValue(Field.YEAR, "2004");
        return result;
    }

    public void setHillclimbMetric(SelectedTag newType) {
        if (newType.getTags() == TAGS_METRIC) {
            m_hillclimbMetric = newType.getSelectedTag().getID();
        }
    }

    public double getInternalHillclimbSetRatio() {
        return m_RatioInternalHillclimbSet;
    }

    public void setInternalHillclimbSetRatio(double v) {
        m_RatioInternalHillclimbSet = v;
    }

    public void setExternalHillclimbSet(Instances hillclimb) {
        m_ExternalHillclimbSet = hillclimb;
    }

    private Instances[] splitData(Instances data, double ratio) throws Exception {
        Instances[] datasets = new Instances[2];
        datasets[0] = new Instances(data, 0);
        datasets[1] = new Instances(data, 0);
        int cutpoint = (int) (ratio * data.numInstances());
        for (int i = 0; i < data.numInstances(); i++) {
            Instance inst = data.instance(i);
            if (i > cutpoint) {
                datasets[0].add(inst);
            } else {
                datasets[1].add(inst);
            }
        }
        return datasets;
    }

    public void setBaseClassifierLibrary(BESBaseClassifierLib lib) {
        m_BCLibrary = lib;
    }

    public int getHillclimbIterations() {
        return m_hillclimbIterations;
    }

    public void setHillclimbIterations(int n) throws Exception {
        if (n < 0) {
            throw new IllegalArgumentException(
                    "EnsembleSelection: Number of hillclimb iterations "
                    + "must be positive.");
        }
        m_hillclimbIterations = n;
    }

    public void buildClassifier(Instances data) throws Exception {

        if (m_Debug) {
            System.out.println("Build classifier ...");
            System.out.println("metric: " + m_hillclimbMetric);
        }
        Random rand = new Random(m_Seed);
        if (m_RandomizeDatset) {
            data.randomize(rand);
        }

        Instances buildset = null;
        Instances hillclimbset = null;

        if (this.m_ExternalHillclimbSet != null) {
            buildset = data;
            hillclimbset = m_ExternalHillclimbSet;
            if (m_Debug) {
                System.out.println("external hillclimb set size: " + hillclimbset.numInstances());
            }
        } else {
            Instances[] datasets = splitData(data, m_RatioInternalHillclimbSet);
            data = null;
            buildset = datasets[0];
            hillclimbset = datasets[1];

            if (m_Debug) {
                System.out.println("internal hillclimb set size: " + hillclimbset.numInstances());
            }
        }

        if (m_Debug) {
            System.out.println("build set size: " + buildset.numInstances());
            System.out.println("hillclimb set size: " + hillclimbset.numInstances());
        }

        if (m_BCLibrary == null || m_BCLibrary.size() == 0) {
            throw new Exception("Base classifier has not been set or no "
                    + "classifier has been added to the library");
        }

        m_BCLibrary.trainAll(buildset, hillclimbset);

        if (m_Debug) {
            System.out.println("lib size (trained): " + m_BCLibrary.size());
            Calendar cal = Calendar.getInstance();
            System.out.printf("Current time and date: %tc\n", cal);
        }

        double[][][] hillClimbPredictions = m_BCLibrary.getHillclimbPredictions();

        weka.classifiers.meta.baggingEnsembleSelection.BESModelBag modelBag =
                new weka.classifiers.meta.baggingEnsembleSelection.BESModelBag(hillClimbPredictions,
                1.0,
                m_Debug);

        int modelWeights[] = new int[m_BCLibrary.size()];


        if (this.m_UseBestModel) {
            weka.classifiers.meta.baggingEnsembleSelection.BESModelBag model_bag = new weka.classifiers.meta.baggingEnsembleSelection.BESModelBag(hillClimbPredictions, 1.0, m_Debug);
            int[] modelPicked = model_bag.sortInitialize(1, false, hillclimbset, m_hillclimbMetric);
            // Then give it a weight of 1, while all others remain 0.
            modelWeights[modelPicked[0]] = 1;
            m_chosen_models = new BESBaseClassifier[1];
            m_chosen_models[0] = m_BCLibrary.get(modelPicked[0]);
            m_chosen_model_weights = new int[1];
            m_chosen_model_weights[0] = 1;
            m_total_weight = 1;
        } else {
            
            for (int i = 0; i < m_numModelBags; i++) {
                // And shuffle it.
                modelBag.shuffle(rand);
                // Sort initialize, if the ratio greater than 0.
                boolean useGreedy = true;
                modelBag.sortInitialize(m_BCLibrary.size(),
                        useGreedy,
                        hillclimbset,
                        m_hillclimbMetric);

                boolean withReplacement = false;
                for (int j = 0; j < getHillclimbIterations(); ++j) {
                    modelBag.forwardSelect(withReplacement, hillclimbset, m_hillclimbMetric);

                }
                int[] bagWeights = modelBag.getModelWeights();
                for (int j = 0; j < bagWeights.length; ++j) {
                    modelWeights[j] += bagWeights[j];
                }
            }

            int libraryIndex = 0;
            int chosenModels = 0;

            //for (int i = 0; i < bagWeights.length; i++) {
            for (int i = 0; i < m_BCLibrary.size(); i++) { // m_BCLibrary
                int weightOfModel = modelWeights[libraryIndex++];
                m_total_weight += weightOfModel;
                if (weightOfModel > 0) {
                    // If the model was chosen at least once, increment the
                    // number of chosen models.
                    ++chosenModels;
                }
            }

            m_chosen_models = new BESBaseClassifier[chosenModels];
            m_chosen_model_weights = new int[chosenModels];

            libraryIndex = 0;
            int chosenIndex = 0;

            for (int i = 0; i < this.m_BCLibrary.size(); i++) {
                int weightOfModel = modelWeights[i];
                BESBaseClassifier bc = m_BCLibrary.get(i);
                if (weightOfModel > 0) {
                    if (m_Debug) {
                        System.out.println(bc.getWekaClassifier().getClass().getSimpleName());
                    }
                    m_chosen_models[chosenIndex] = bc;
                    m_chosen_model_weights[chosenIndex] = weightOfModel;

                    ++chosenIndex;
                }
            }
        }

        m_FinalEnsembleSize = m_chosen_models.length;
        
        m_bestPerformance = modelBag.getFinalBestPerformance();

        if (m_Debug) {
            System.out.println("Model built, final ensemble size: " + m_chosen_models.length + ", m_bestPerformance: " + m_bestPerformance);
        }
    }

    public double getBestPerformance() {
        return m_bestPerformance;
    }
    
    public double[] distributionForInstance(Instance instance) throws Exception {

        double cachedPreds[][] = null;
        double[] prediction = new double[instance.numClasses()];
        for (int i = 0; i < prediction.length; ++i) {
            prediction[i] = 0.0;
        }

        for (int i = 0; i < m_chosen_models.length; i++) {
            double[] predictionForThisModel = null;
            if (cachedPreds == null) {
                predictionForThisModel = m_chosen_models[i].getWekaClassifier().distributionForInstance(instance);
            }
            if (predictionForThisModel != null) {
                for (int j = 0; j < prediction.length; ++j) {
                    prediction[j] += m_chosen_model_weights[i] * predictionForThisModel[j] / m_total_weight;
                }
            }
        }

        // normalize to add up to 1.
        if (instance.classAttribute().isNominal()) {
            if (Utils.sum(prediction) > 0) {
                Utils.normalize(prediction);
            }
        }
        return prediction;
    }
}
