/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESEnsembleLibraryModel.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta.baggingEnsembleSelection;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.MethodDescriptor;
import java.beans.PropertyDescriptor;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Comparator;
import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.RevisionHandler;
import weka.core.RevisionUtils;
import weka.core.Utils;
import weka.gui.GenericObjectEditor;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 * @author Cornell University
 */
public class BESEnsembleLibraryModel
        implements Serializable, RevisionHandler {

    private static final long serialVersionUID = 7932816660173443200L;
    
    protected Classifier m_Classifier;
    
    protected String m_DescriptionText;
    
    protected String m_ErrorText;
    
    protected boolean m_OptionsWereValid;
    
    protected String m_StringRepresentation;

    public BESEnsembleLibraryModel() {
        super();
    }

    public BESEnsembleLibraryModel(Classifier classifier) {

        m_Classifier = classifier;

        m_StringRepresentation = toString();

        updateDescriptionText();
    }

    public void testOptions() {

        Classifier testClassifier = null;
        try {
            testClassifier = ((Classifier) m_Classifier.getClass().newInstance());
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        }

        setOptionsWereValid(true);
        setErrorText(null);
        updateDescriptionText();

        try {
            ((OptionHandler) testClassifier).setOptions(((OptionHandler) m_Classifier).getOptions());
        } catch (Exception e) {

            setOptionsWereValid(false);
            setErrorText(e.getMessage());
        }

        updateDescriptionText();
    }

    public Classifier getClassifier() {
        return m_Classifier;
    }

    public String getStringRepresentation() {
        return m_StringRepresentation;
    }

    public void setDescriptionText(String descriptionText) {
        m_DescriptionText = descriptionText;
    }

    public String getDescriptionText() {
        return m_DescriptionText;
    }

    public void setErrorText(String errorText) {
        this.m_ErrorText = errorText;
    }

    public String getErrorText() {
        return m_ErrorText;
    }

    public void setOptionsWereValid(boolean optionsWereValid) {
        this.m_OptionsWereValid = optionsWereValid;
    }

    public boolean getOptionsWereValid() {
        return m_OptionsWereValid;
    }

    public String toString() {
        String str = m_Classifier.getClass().getName();
        str += " " + Utils.joinOptions(((OptionHandler) m_Classifier).getOptions());
        return str;
    }

    public Class getModelClass() {
        return m_Classifier.getClass();
    }

    public String[] getOptions() {
        return ((OptionHandler) m_Classifier).getOptions();
    }

    public void updateDescriptionText() {

        String toolTipText = new String("<html>");

        if (!m_OptionsWereValid) {

            toolTipText += "<font COLOR=\"#FF0000\">"
                    + "<b>Invalid Model:</b><br>" + m_ErrorText + "<br></font>";
        }

        toolTipText += "<TABLE>";

        PropertyDescriptor properties[] = null;
        MethodDescriptor methods[] = null;

        try {
            BeanInfo bi = Introspector.getBeanInfo(m_Classifier.getClass());
            properties = bi.getPropertyDescriptors();
            methods = bi.getMethodDescriptors();
        } catch (IntrospectionException ex) {
            System.err.println("LibraryModel: Couldn't introspect");
            return;
        }

        for (int i = 0; i < properties.length; i++) {

            if (properties[i].isHidden() || properties[i].isExpert()) {
                continue;
            }

            String name = properties[i].getDisplayName();
            Class type = properties[i].getPropertyType();
            Method getter = properties[i].getReadMethod();
            Method setter = properties[i].getWriteMethod();

            // Only display read/write properties.
            if (getter == null || setter == null) {
                continue;
            }

            try {
                Object args[] = {};
                Object value = getter.invoke(m_Classifier, args);

                PropertyEditor editor = null;
                Class pec = properties[i].getPropertyEditorClass();
                if (pec != null) {
                    try {
                        editor = (PropertyEditor) pec.newInstance();
                    } catch (Exception ex) {
                        // Drop through.
                    }
                }
                if (editor == null) {
                    editor = PropertyEditorManager.findEditor(type);
                }
                //editors[i] = editor;

                // If we can't edit this component, skip it.
                if (editor == null) {
                    continue;
                }
                if (editor instanceof GenericObjectEditor) {
                    ((GenericObjectEditor) editor).setClassType(type);
                }

                // Don't try to set null values:
                if (value == null) {
                    continue;
                }

                toolTipText += "<TR><TD>" + name + "</TD><TD>"
                        + value.toString() + "</TD></TR>";

            } catch (InvocationTargetException ex) {
                System.err.println("Skipping property " + name
                        + " ; exception on target: " + ex.getTargetException());
                ex.getTargetException().printStackTrace();
                continue;
            } catch (Exception ex) {
                System.err.println("Skipping property " + name
                        + " ; exception: " + ex);
                ex.printStackTrace();
                continue;
            }
        }

        toolTipText += "</TABLE>";
        toolTipText += "</html>";
        m_DescriptionText = toolTipText;
    }

    public String getRevision() {
        return RevisionUtils.extract("$Revision: 6041 $");
    }

    public final static class BESEnsembleLibraryModelComparator
            implements Comparator, Serializable, RevisionHandler {

        private static final long serialVersionUID = -6522464740036141188L;

        public int compare(Object o1, Object o2) {

            int comparison = 0;

            if (o1 instanceof BESEnsembleLibraryModel
                    && o2 instanceof BESEnsembleLibraryModel) {

                comparison = ((String) ((BESEnsembleLibraryModel) o1)
                        .getStringRepresentation())
                        .compareTo(((String) ((BESEnsembleLibraryModel) o2)
                        .getStringRepresentation()));

            }

            return comparison;
        }

        public String getRevision() {
            return RevisionUtils.extract("$Revision: 6041 $");
        }
    }
}
