/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESEnsembleSelectionLibraryModel.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */
package weka.classifiers.meta.baggingEnsembleSelection;

import weka.classifiers.Classifier;
import weka.classifiers.AbstractClassifier;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.RevisionUtils;
import weka.core.Utils;
import weka.core.OptionHandler;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.zip.Adler32;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 * @author Cornell University
 */
public class BESEnsembleSelectionLibraryModel
        extends BESEnsembleLibraryModel
        implements Serializable {

    private static final long serialVersionUID = -6426075459862947640L;
    /** The default file extension for ensemble library models */
    public static final String FILE_EXTENSION = ".elm";
    /** the models */
    private Classifier[] m_models = null;
    /** The seed that was used to create this model */
    private int m_seed;
    /**
     * The checksum of the instances.arff object that was used to create this
     * model
     */
    private String m_checksum;
    /** The validation ratio that was used to create this model */
    private double m_validationRatio;
    /**
     * The number of folds, or number of CV models that was used to create this
     * "model"
     */
    private int m_folds;
    /**
     * The .elm file name that this model should be saved/loaded to/from
     */
    private String m_fileName;
    /**
     * The debug flag as propagated from the main EnsembleSelection class.
     */
    public transient boolean m_Debug = true;
    /**
     * the validation predictions of this model. First index for the instance.
     * third is for the class (we use distributionForInstance).
     */
    private double[][] m_validationPredictions = null; // = new double[0][0];

    /**
     * Default Constructor
     */
    public BESEnsembleSelectionLibraryModel() {
        //
    }

    /**
     * Constructor for LibaryModel
     *
     * @param classifier		the classifier to use
     * @param seed		the random seed value
     * @param checksum		the checksum
     * @param validationRatio	the ration to use
     * @param folds		the number of folds to use
     */
    public BESEnsembleSelectionLibraryModel(Classifier classifier, int seed,
            String checksum, double validationRatio, int folds) {

        super(classifier);

        m_seed = seed;
        m_checksum = checksum;
        m_validationRatio = validationRatio;
        m_models = null;
        m_folds = folds;
    }

    /**
     * This is used to propagate the m_Debug flag of the EnsembleSelection
     * classifier to this class. There are things we would want to print out
     * here also.
     *
     * @param debug	if true additional information is output
     */
    public void setDebug(boolean debug) {
        m_Debug = debug;
    }

    /**
     * Returns the average of the prediction of the models across all folds.
     *
     * @param instance	the instance to get predictions for
     * @return		the average prediction
     * @throws Exception	if something goes wrong
     */
    public double[] getAveragePrediction(Instance instance) throws Exception {

        // Return the average prediction from all classifiers that make up
        // this model.
        double average[] = new double[instance.numClasses()];
        for (int i = 0; i < m_folds; ++i) {
            // Some models alter the instance (MultiLayerPerceptron), so we need
            // to copy it.
            Instance temp_instance = (Instance) instance.copy();
            double[] pred = getFoldPrediction(temp_instance, i);
            if (pred == null) {
                // Some models have bugs whereby they can return a null
                // prediction
                // array (again, MultiLayerPerceptron). We return null, and this
                // should be handled above in EnsembleSelection.
                System.err.println("Null validation predictions given: "
                        + getStringRepresentation());
                return null;
            }
            if (i == 0) {
                // The first time through the loop, just use the first returned
                // prediction array. Just a simple optimization.
                average = pred;
            } else {
                // For the rest, add the prediction to the average array.
                for (int j = 0; j < pred.length; ++j) {
                    average[j] += pred[j];
                }
            }
        }
        if (instance.classAttribute().isNominal()) {
            // Normalize predictions for classes to add up to 1.
            Utils.normalize(average);
        } else {
            average[0] /= m_folds;
        }
        return average;
    }

    /**
     * Basic Constructor
     *
     * @param classifier	the classifier to use
     */
    public BESEnsembleSelectionLibraryModel(Classifier classifier) {
        super(classifier);
    }

    /**
     * Returns prediction of the classifier for the specified fold.
     *
     * @param instance
     *            instance for which to make a prediction.
     * @param fold
     *            fold number of the classifier to use.
     * @return the prediction for the classes
     * @throws Exception if prediction fails
     */
    public double[] getFoldPrediction(Instance instance, int fold)
            throws Exception {

        return m_models[fold].distributionForInstance(instance);
    }

    /**
     * Creates the model. If there are n folds, it constructs n classifiers
     * using the current Classifier class and options. If the model has already
     * been created or loaded, starts fresh.
     *
     * @param data		the data to work with
     * @param hillclimbData	the data for hillclimbing
     * @param dataDirectoryName	the directory to use
     * @param algorithm		the type of algorithm
     * @throws Exception		if something goeds wrong
     */
    public void createModel(Instances[] data, Instances[] hillclimbData,
            String dataDirectoryName, int algorithm) throws Exception {

        String modelFileName = getFileName(getStringRepresentation());

        File modelFile = new File(dataDirectoryName, modelFileName);

        String relativePath = (new File(dataDirectoryName)).getName()
                + File.separatorChar + modelFileName;
        // if (m_Debug) System.out.println("setting relative path to:
        // "+relativePath);
        setFileName(relativePath);

        if (!modelFile.exists()) {

            Date startTime = new Date();

            String lockFileName = BESEnsembleSelectionLibraryModel.getFileName(getStringRepresentation());
            lockFileName = lockFileName.substring(0, lockFileName.length() - 3)
                    + "LCK";
            File lockFile = new File(dataDirectoryName, lockFileName);

            if (lockFile.exists()) {
                if (m_Debug) {
                    System.out.println("Detected lock file.  Skipping: "
                            + lockFileName);
                }
                throw new Exception("Lock File Detected: " + lockFile.getName());

            } else { // if (algorithm ==
                // EnsembleSelection.ALGORITHM_BUILD_LIBRARY) {
                // This lock file lets other computers that might be sharing the
                // same file
                // system that this model is already being trained so they know
                // to move ahead
                // and train other models.

                if (lockFile.createNewFile()) {

                    if (m_Debug) {
                        System.out.println("lock file created: " + lockFileName);
                    }

                    if (m_Debug) {
                        System.out.println("Creating model in locked mode: "
                                + modelFile.getPath());
                    }

                    m_models = new Classifier[m_folds];
                    for (int i = 0; i < m_folds; ++i) {

                        try {
                            m_models[i] = AbstractClassifier.forName(getModelClass().getName(), null);
                            ((OptionHandler) m_models[i]).setOptions(getOptions());
                        } catch (Exception e) {
                            throw new Exception("Invalid Options: "
                                    + e.getMessage());
                        }
                    }

                    try {
                        for (int i = 0; i < m_folds; ++i) {
                            train(data[i], i);
                        }
                    } catch (Exception e) {
                        throw new Exception("Could not Train: "
                                + e.getMessage());
                    }

                    Date endTime = new Date();
                    int diff = (int) (endTime.getTime() - startTime.getTime());

                    if (m_Debug) {
                        System.out.println("Train time for " + modelFileName
                                + " was: " + diff);
                    }

                    if (m_Debug) {
                        System.out.println("Generating validation set predictions");
                    }

                    startTime = new Date();

                    int total = 0;
                    for (int i = 0; i < m_folds; ++i) {
                        total += hillclimbData[i].numInstances();
                    }

                    m_validationPredictions = new double[total][];

                    int preds_index = 0;
                    for (int i = 0; i < m_folds; ++i) {
                        for (int j = 0; j < hillclimbData[i].numInstances(); ++j) {
                            Instance temp = (Instance) hillclimbData[i].instance(j).copy();// new
                            // Instance(m_hillclimbData[i].instance(j));
                            // must copy the instance because SOME classifiers
                            // (I'm not pointing fingers...
                            // MULTILAYERPERCEPTRON)
                            // change the instance!

                            m_validationPredictions[preds_index] = getFoldPrediction(
                                    temp, i);

                            if (m_validationPredictions[preds_index] == null) {
                                throw new Exception(
                                        "Null validation predictions given: "
                                        + getStringRepresentation());
                            }

                            ++preds_index;
                        }
                    }

                    endTime = new Date();
                    diff = (int) (endTime.getTime() - startTime.getTime());

                    // if (m_Debug) System.out.println("Generated a validation
                    // set array of size: "+m_validationPredictions.length);
                    if (m_Debug) {
                        System.out.println("Time to create validation predictions was: "
                                + diff);
                    }

                    BESEnsembleSelectionLibraryModel.saveModel(dataDirectoryName,
                            this);

                    if (m_Debug) {
                        System.out.println("deleting lock file: "
                                + lockFileName);
                    }
                    lockFile.delete();

                } else {

                    if (m_Debug) {
                        System.out.println("Could not create lock file.  Skipping: "
                                + lockFileName);
                    }
                    throw new Exception(
                            "Could not create lock file.  Skipping: "
                            + lockFile.getName());

                }

            }

        } else {
            // This branch is responsible for loading a model from a .elm file

            if (m_Debug) {
                System.out.println("Loading model: " + modelFile.getPath());
            }
            // now we need to check to see if the model is valid, if so then
            // load it
            Date startTime = new Date();

            BESEnsembleSelectionLibraryModel newModel = loadModel(modelFile.getPath());

            if (!newModel.getStringRepresentation().equals(
                    getStringRepresentation())) {
                throw new Exception(
                        "String representations "
                        + newModel.getStringRepresentation() + " and "
                        + getStringRepresentation() + " not equal");
            }

            if (!newModel.getChecksum().equals(getChecksum())) {
                throw new Exception("Checksums "
                        + newModel.getChecksum() + " and " + getChecksum()
                        + " not equal");
            }

            if (newModel.getSeed() != getSeed()) {
                throw new Exception("Seeds "
                        + newModel.getSeed() + " and " + getSeed()
                        + " not equal");
            }

            if (newModel.getFolds() != getFolds()) {
                throw new Exception("Folds "
                        + newModel.getFolds() + " and " + getFolds()
                        + " not equal");
            }

            if (newModel.getValidationRatio() != getValidationRatio()) {
                throw new Exception("Validation Ratios "
                        + newModel.getValidationRatio() + " and "
                        + getValidationRatio() + " not equal");
            }

            // setFileName(modelFileName);

            m_models = newModel.getModels();
            m_validationPredictions = newModel.getValidationPredictions();

            Date endTime = new Date();
            int diff = (int) (endTime.getTime() - startTime.getTime());
            if (m_Debug) {
                System.out.println("Time to load " + modelFileName + " was: "
                        + diff);
            }
        }
    }

    /**
     * The purpose of this method is to "rehydrate" the classifier object fot
     * this library model from the filesystem.
     *
     * @param workingDirectory	the working directory to use
     */
    public void rehydrateModel(String workingDirectory) {

        if (m_models == null) {

            File file = new File(workingDirectory, m_fileName);

            if (m_Debug) {
                System.out.println("Rehydrating Model: " + file.getPath());
            }
            BESEnsembleSelectionLibraryModel model = BESEnsembleSelectionLibraryModel.loadModel(file.getPath());

            m_models = model.getModels();

        }
    }

    /**
     * Releases the model from memory. TODO - need to be saving these so we can
     * retrieve them later!!
     */
    public void releaseModel() {
        /*
         * if (m_unsaved) { saveModel(); }
         */
        m_models = null;
    }

    /**
     * Train the classifier for the specified fold on the given data
     *
     * @param trainData	the data to train with
     * @param fold	the fold number
     * @throws Exception	if something goes wrong, e.g., out of memory
     */
    public void train(Instances trainData, int fold) throws Exception {
        if (m_models != null) {

            try {
                // OK, this is it... this is the point where our code surrenders
                // to the weka classifiers.
                m_models[fold].buildClassifier(trainData);
            } catch (Throwable t) {
                m_models[fold] = null;
                throw new Exception(
                        "Exception caught while training: (null could mean out of memory)"
                        + t.getMessage());
            }

        } else {
            throw new Exception("Cannot train: model was null");
            // TODO: throw Exception?
        }
    }

    /**
     * Set the seed
     *
     * @param seed	the seed value
     */
    public void setSeed(int seed) {
        m_seed = seed;
    }

    /**
     * Get the seed
     *
     * @return the seed value
     */
    public int getSeed() {
        return m_seed;
    }

    /**
     * Sets the validation set ratio (only meaningful if folds == 1)
     *
     * @param validationRatio	the new ration
     */
    public void setValidationRatio(double validationRatio) {
        m_validationRatio = validationRatio;
    }

    /**
     * get validationRatio
     *
     * @return		the current ratio
     */
    public double getValidationRatio() {
        return m_validationRatio;
    }

    /**
     * Set the number of folds for cross validation. The number of folds also
     * indicates how many classifiers will be built to represent this model.
     *
     * @param folds	the number of folds to use
     */
    public void setFolds(int folds) {
        m_folds = folds;
    }

    /**
     * get the number of folds
     *
     * @return		the current number of folds
     */
    public int getFolds() {
        return m_folds;
    }

    /**
     * set the checksum
     *
     * @param instancesChecksum	the new checksum
     */
    public void setChecksum(String instancesChecksum) {
        m_checksum = instancesChecksum;
    }

    /**
     * get the checksum
     *
     * @return		the current checksum
     */
    public String getChecksum() {
        return m_checksum;
    }

    /**
     * Returs the array of classifiers
     *
     * @return		the current models
     */
    public Classifier[] getModels() {
        return m_models;
    }

    /**
     * Sets the .elm file name for this library model
     *
     * @param fileName	the new filename
     */
    public void setFileName(String fileName) {
        m_fileName = fileName;
    }

    /**
     * Gets a checksum for the string defining this classifier. This is used to
     * preserve uniqueness in the classifier names.
     *
     * @param string	the classifier definition
     * @return		the checksum string
     */
    public static String getStringChecksum(String string) {

        String checksumString = null;

        try {

            Adler32 checkSummer = new Adler32();

            byte[] utf8 = string.toString().getBytes("UTF8");
            ;

            checkSummer.update(utf8);
            checksumString = Long.toHexString(checkSummer.getValue());

        } catch (UnsupportedEncodingException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        return checksumString;
    }

    /**
     * The purpose of this method is to get an appropriate file name for a model
     * based on its string representation of a model. All generated filenames
     * are limited to less than 128 characters and all of them will end with a
     * 64 bit checksum value of their string representation to try to maintain
     * some uniqueness of file names.
     *
     * @param stringRepresentation	string representation of model
     * @return				unique filename
     */
    public static String getFileName(String stringRepresentation) {

        // Get rid of space and quote marks(windows doesn't lke them)
        String fileName = stringRepresentation.trim().replace(' ', '_').replace('"', '_');

        if (fileName.length() > 115) {

            fileName = fileName.substring(0, 115);

        }

        fileName += getStringChecksum(stringRepresentation)
                + BESEnsembleSelectionLibraryModel.FILE_EXTENSION;

        return fileName;
    }

    /**
     * Saves the given model to the specified file.
     *
     * @param directory	the directory to save the model to
     * @param model	the model to save
     */
    public static void saveModel(String directory,
            BESEnsembleSelectionLibraryModel model) {

        try {
            String fileName = getFileName(model.getStringRepresentation());

            File file = new File(directory, fileName);

            // System.out.println("Saving model: "+file.getPath());

            // model.setFileName(new String(file.getPath()));

            // Serialize to a file
            ObjectOutput out = new ObjectOutputStream(
                    new FileOutputStream(file));
            out.writeObject(model);

            out.close();

        } catch (IOException e) {

            e.printStackTrace();
        }
    }

    /**
     * loads the specified model
     *
     * @param modelFilePath	the path of the model
     * @return			the model
     */
    public static BESEnsembleSelectionLibraryModel loadModel(String modelFilePath) {

        BESEnsembleSelectionLibraryModel model = null;

        try {

            File file = new File(modelFilePath);

            ObjectInputStream in = new ObjectInputStream(new FileInputStream(
                    file));

            model = (BESEnsembleSelectionLibraryModel) in.readObject();

            in.close();

        } catch (ClassNotFoundException e) {

            e.printStackTrace();

        } catch (IOException e) {

            e.printStackTrace();

        }

        return model;
    }

    public double[][] getValidationPredictions() {
        return m_validationPredictions;
    }

    /**
     * setter for validation predictions
     *
     * @param predictions	the new validation predictions
     */
    public void setValidationPredictions(double[][] predictions) {
        if (m_Debug) {
            System.out.println("Saving validation array of size "
                    + predictions.length);
        }
        m_validationPredictions = new double[predictions.length][];
        System.arraycopy(predictions, 0, m_validationPredictions, 0,
                predictions.length);
    }

    /**
     * Returns the revision string.
     *
     * @return		the revision
     */
    public String getRevision() {
        return RevisionUtils.extract("$Revision: 5928 $");
    }
}
