/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESHelper.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta.baggingEnsembleSelection;

import java.io.Serializable;
import weka.classifiers.Classifier;
import weka.core.Instances;

import weka.classifiers.*;
import weka.core.*;
import java.io.Serializable;
import java.util.ArrayList;

import java.io.Serializable;
import java.util.*;
import weka.classifiers.meta.baggingEnsembleSelection.BESHelper.BESBaseClassifier;
import weka.core.*;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 */
public class BESHelper {

    /**
     * metric: Correlation coefficient
     */
    public static final int METRIC_CC = 0;
    /**
     * metric: RMSE
     */
    public static final int METRIC_RMSE = 1;
    /**
     * metric: ROC
     */
    public static final int METRIC_ROC = 2;
    /**
     * metric: Precision
     */
    public static final int METRIC_PRECISION = 3;
    /**
     * metric: Recall
     */
    public static final int METRIC_RECALL = 4;
    /**
     * metric: FScore
     */
    public static final int METRIC_FSCORE = 5;
    /**
     * metric: All
     */
    public static final int METRIC_ALL = 6;
    /**
     * metric: Weighted true positive
     */
    public static final int METRIC_WEIGHTEDTP = 7;
    /**
     * metric: Mean abs error
     */
    public static final int METRIC_MAE = 8;
    /**
     * metric: Accuracy
     */
    public static final int METRIC_ACCURACY = 9;

    public static double getMetric(Evaluation eval, int metric_index) throws Exception {
        //
        switch (metric_index) {
            case METRIC_ACCURACY:
                return eval.pctCorrect();
            case METRIC_WEIGHTEDTP:
                return eval.weightedTruePositiveRate();
            case METRIC_CC:
                return eval.correlationCoefficient();
            case METRIC_RMSE:
                return -1.0 * eval.rootMeanSquaredError();
            case METRIC_MAE:
                return -1.0 * eval.meanAbsoluteError();
            case METRIC_ROC:
                return eval.weightedAreaUnderROC(); //TODO - is 1 right?
            case METRIC_PRECISION:
                return eval.precision(1); //TODO - same question
            case METRIC_RECALL:
                return eval.recall(1); //TODO - same question
            case METRIC_FSCORE:
                return eval.fMeasure(1); //TODO - same question
            case METRIC_ALL:
                double average = 0;
                int num_metrics = 0;
                average += eval.pctCorrect();
                ++num_metrics;
                average += 1.0 - eval.rootMeanSquaredError();
                ++num_metrics;
                average += eval.areaUnderROC(1);
                ++num_metrics;
                average += eval.precision(1);
                ++num_metrics;
                average += eval.recall(1);
                ++num_metrics;
                average += eval.fMeasure(1);
                ++num_metrics;
                return average / num_metrics;
            default:
                return METRIC_RMSE;
        }
    }
    
    public static class BESBaseClassifierLib implements Serializable {

        private double[][][] m_predictions;
        private ArrayList<BESBaseClassifier> m_Pool = new ArrayList<BESBaseClassifier>();
        private boolean m_Debug = true;

        public void setDebug(boolean d) {
            m_Debug = d;
        }

        public void add(BESBaseClassifier algo) {
            algo.setDebug(m_Debug);
            m_Pool.add(algo);
        }

        public int size() {
            return m_Pool.size();
        }

        public void trainAll(Instances buildSet, Instances hillclimbSet) throws Exception {

            m_predictions = new double[size()][hillclimbSet.numInstances()][buildSet.numClasses()];

            for (Object baseClassifier : m_Pool) {
                BESBaseClassifier bc = (BESBaseClassifier) baseClassifier;
                try {
                    bc.train(buildSet);
                    //TODO: double check this line
                    if (bc.hasBuildingError()) {
                        throw new Exception("We don't want this to happen, becoz I haven't done that part...");
                        //m_Pool.remove(i);
                        //i--;
                    }
                } catch (Exception e) {
                    System.err.println("BaseClassifierLibrary.TrainAll Err : " + e.toString());
                }
            }

            for (int i = 0; i < hillclimbSet.numInstances(); i++) {
                Instance inst = hillclimbSet.instance(i);
                for (int j = 0; j < m_Pool.size(); j++) {
                    BESBaseClassifier bc = m_Pool.get(j);
                    double preds[] = bc.getWekaClassifier().distributionForInstance(inst);
                    m_predictions[j][i] = preds;
                }
            }
        }

        public double[][][] getHillclimbPredictions() {
            return m_predictions;
        }

        public BESBaseClassifier get(int i) {
            BESBaseClassifier bc = m_Pool.get(i);
            return bc;
        }
    }

    public static class BESBaseClassifier implements Serializable {

        private Classifier m_BaseClassifier = null;
        private boolean m_HasBuildingErr = false;
        private boolean m_Debug = false;
        private String m_ThreadID = "unkown";
        private boolean m_ReTrainFull = false;

        public void setDebug(boolean d) {
            m_Debug = d;
        }

        public BESBaseClassifier(String id, Classifier wekaClassifier) {
            if (!id.isEmpty()) {
                m_ThreadID = id;
            }
            m_BaseClassifier = wekaClassifier;
        }

        public void train(Instances data) throws Exception {
            try {
                m_BaseClassifier.buildClassifier(data);
                if (m_Debug) {
                    System.err.println(m_ThreadID + " done!");
                }
            } catch (Exception e) {
                System.err.println(m_ThreadID + ": " + e.toString());
                m_HasBuildingErr = true;
            }
        }

        public void retrainFull(Instances data) throws Exception {
            try {
                m_BaseClassifier.buildClassifier(data);
                if (m_Debug) {
                    System.err.println(m_ThreadID + " RetrainFull done!");
                }
                m_ReTrainFull = true;
            } catch (Exception e) {
                System.err.println(e.toString());
                m_HasBuildingErr = true;
            }
        }

        public boolean isReTrainFullDone() {
            return m_ReTrainFull;
        }

        public Classifier getWekaClassifier() {
            return m_BaseClassifier;
        }

        public boolean hasBuildingError() {
            return m_HasBuildingErr;
        }
    }
}
