/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESModelBag.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta.baggingEnsembleSelection;

import weka.classifiers.Evaluation;
import weka.core.Instances;
import weka.core.RevisionHandler;
import weka.core.RevisionUtils;

import java.util.Random;

/**
 *
 * @author Quan Sun (quan.sun.nz@gmail.com)
 * @author University of Waikato
 * @author Cornell University
 */
public class BESModelBag
        implements RevisionHandler {

    private double m_models[][][];
    private int m_modelIndex[];
    private int m_bagSize;
    private int m_numChosen;
    private int m_timesChosen[];
    private boolean m_debug;
    private double m_bestPerformance;
    private int m_bestTimesChosen[];

    public BESModelBag(double models[][][], double bag_percent, boolean debug) {
        m_debug = debug;
        if (models.length == 0) {
            throw new IllegalArgumentException(
                    "ModelBag needs at least 1 model.");
        }
        m_bagSize = (int) ((double) models.length * bag_percent);
        m_models = models;
        m_modelIndex = new int[m_models.length];
        m_timesChosen = new int[m_models.length];
        m_bestTimesChosen = m_timesChosen;
        m_bestPerformance = Double.NEGATIVE_INFINITY;
        // Initially, no models are chosen.
        m_numChosen = 0;
        // Prepare our virtual indexing scheme. Initially, the indexes are
        // the same as the original.
        for (int i = 0; i < m_models.length; ++i) {
            m_modelIndex[i] = i;
            m_timesChosen[i] = 0;
        }
    }

    /**
     * Swap model at virtual index i with model at virtual index j. This is used
     * to shuffle the models. We do not change m_models, only the arrays which
     * use the virtual indexing; m_modelIndex and m_timesChosen.
     *
     * @param i	first index
     * @param j	second index
     */
    private void swap(int i, int j) {
        if (i != j) {
            int temp_index = m_modelIndex[i];
            m_modelIndex[i] = m_modelIndex[j];
            m_modelIndex[j] = temp_index;

            int tempWeight = m_timesChosen[i];
            m_timesChosen[i] = m_timesChosen[j];
            m_timesChosen[j] = tempWeight;
        }
    }

    /**
     * Shuffle the models. The order in m_models is preserved, but we change our
     * virtual indexes around.
     *
     * @param rand	the random number generator to use
     */
    public void shuffle(Random rand) {
        if (m_models.length < 2) {
            return;
        }

        for (int i = 0; i < m_models.length; ++i) {
            int swap_index = rand.nextInt(m_models.length - 1);
            if (swap_index >= i) {
                ++swap_index; // don't swap with itself
            }
            swap(i, swap_index);
        }
    }

    /**
     * Convert an array of weights using virtual indices to an array of weights
     * using real indices.
     *
     * @param virtual_weights	the virtual indices
     * @return	the real indices
     */
    private int[] virtualToRealWeights(int virtual_weights[]) {
        int real_weights[] = new int[virtual_weights.length];
        for (int i = 0; i < real_weights.length; ++i) {
            real_weights[m_modelIndex[i]] = virtual_weights[i];
        }
        return real_weights;
    }

    /**
     *
     */
    private void updateBestTimesChosen() {
        m_bestTimesChosen = virtualToRealWeights(m_timesChosen);
    }

    /**
     * Sort initialize the bag.
     *
     * @param num the Maximum number of models to initialize with
     * @param greedy True if we do greedy addition, up to num. Greedy sort
     * initialization adds models (up to num) in order of best to worst
     * performance until performance no longer improves.
     * @param instances the data set (needed for performance evaluation)
     * @param metric metric for which to optimize. See EnsembleMetricHelper
     * @return returns an array of indexes which were selected, in order
     * starting from the model with best performance.
     * @throws Exception if something goes wrong
     */
    public int[] sortInitialize(int num, boolean greedy, Instances instances,
            int metric) throws Exception {

        // First, get the performance of each model
        double performance[] = new double[m_bagSize];
        for (int i = 0; i < m_bagSize; ++i) {
            performance[i] = evaluatePredictions(instances, model(i), metric);
        }
        int bestModels[] = new int[num]; // we'll use this to save model info
        // Now sort the models by their performance... note we only need the
        // first "num",
        // so we don't actually bother to sort the whole thing... instead, we
        // pick the num best
        // by running num iterations of selection sort.
        for (int i = 0; i < num; ++i) {
            int max_index = i;
            double max_value = performance[i];
            for (int j = i + 1; j < m_bagSize; ++j) {
                // Find the best model which we haven't already selected
                if (performance[j] > max_value) {
                    max_value = performance[j];
                    max_index = j;
                }
            }
            // Swap ith model in to the ith position (selection sort)
            this.swap(i, max_index);
            // swap performance numbers, too
            double temp_perf = performance[i];
            performance[i] = performance[max_index];
            performance[max_index] = temp_perf;

            bestModels[i] = m_modelIndex[i];
            if (!greedy) {
                // If we're not being greedy, we just throw the model in
                // no matter what
                ++m_timesChosen[i];
                ++m_numChosen;
            }
        }
        // Now the best "num" models are all sorted and in position.
        if (greedy) {
            // If the "greedy" option was specified, do a smart sort
            // initialization
            // that adds models only so long as they help overall performance.
            // This is what was done in the original Caruana paper.
            double[][] tempPredictions = null;

            double bestPerformance = Double.NEGATIVE_INFINITY;
            // double bestPerformance = 0.0;

            if (num > 0) {
                ++m_timesChosen[0];
                ++m_numChosen;
                updateBestTimesChosen();
            }
            for (int i = 1; i < num; ++i) {
                tempPredictions = computePredictions(i, true);
                double metric_value = evaluatePredictions(instances,
                        tempPredictions, metric);
                if (metric_value > bestPerformance) {
                    // If performance improved, update the appropriate info.
                    bestPerformance = metric_value;

                    m_bestPerformance = bestPerformance;

                    ++m_timesChosen[i];
                    ++m_numChosen;
                    updateBestTimesChosen();
                } else {
                    // We found a model that doesn't help performance, so we
                    // stop adding models.
                    break;
                }
            }
        }
        updateBestTimesChosen();
        if (m_debug) {
            System.out.println("Sort Initialization added best " + m_numChosen
                    + " models to the bag.");
        }
        return bestModels;
    }

    /**
     * Add "weight" to the number of times each model in the bag was chosen.
     * Typically for use with backward elimination.
     *
     * @param weight	the weight to add
     */
    public void weightAll(int weight) {
        for (int i = 0; i < m_bagSize; ++i) {
            m_timesChosen[i] += weight;
            m_numChosen += weight;
        }
        updateBestTimesChosen();
    }

    /**
     * Forward select one model. Will add the model which has the best effect on
     * performance. If replacement is false, and all models are chosen, no
     * action is taken. If a model can be added, one always is (even if it hurts
     * performance).
     *
     * @param withReplacement whether a model can be added more than once.
     * @param instances The dataset, for calculating performance.
     * @param metric The metric to which we will optimize. See
     * EnsembleMetricHelper
     * @throws Exception if something goes wrong
     */
    public void forwardSelect(boolean withReplacement, Instances instances,
            int metric) throws Exception {

        double bestPerformance = Double.NEGATIVE_INFINITY;
        // double bestPerformance = -1.0;

        int bestIndex = -1;
        double tempPredictions[][];

        for (int i = 0; i < m_bagSize; ++i) {
            // For each model in the bag
            if ((m_timesChosen[i] == 0) || withReplacement) {
                // If the model has not been chosen, or we're allowing
                // replacement
                // Get the predictions we would have if we add this model to the
                // ensemble
                tempPredictions = computePredictions(i, true);
                // And find out how the hypothetical ensemble would perform.
                double metric_value = evaluatePredictions(instances,
                        tempPredictions, metric);
                if (metric_value > bestPerformance) {
                    // If it's better than our current best, make it our NEW
                    // best.
                    bestIndex = i;
                    bestPerformance = metric_value;
                }

            }
        }
        if (bestIndex == -1) {
            // Replacement must be false, with more hillclimb iterations than
            // models. Do nothing and return.
            if (m_debug) {
                System.out.println("Couldn't add model.  No action performed.");
            }
            return;
        }
        // We picked bestIndex as our best model. Update appropriate info.
        m_timesChosen[bestIndex]++;
        m_numChosen++;
        if (bestPerformance > m_bestPerformance) {
            // We find the peak of our performance over all hillclimb
            // iterations.
            // If this forwardSelect step improved our overall performance,
            // update
            // our best ensemble info.
            updateBestTimesChosen();
            m_bestPerformance = bestPerformance;
        }
    }

    public double getFinalBestPerformance() {
        //TODO: be careful that METRIC_RMSE, MSE, MAE are less then zero.
        if (m_bestPerformance < 0) {
            return -1 * m_bestPerformance;
        }
        return m_bestPerformance;
    }

    /**
     * returns the model weights
     *
     * @return	the model weights
     */
    public int[] getModelWeights() {
        return m_bestTimesChosen;
    }

    private double[][] model(int index) {
        return m_models[m_modelIndex[index]];
    }

    private double[][] computePredictions(int index_to_change, boolean add) {
        double[][] predictions = new double[m_models[0].length][m_models[0][0].length];
        for (int i = 0; i < m_bagSize; ++i) {
            if (m_timesChosen[i] > 0) {
                for (int j = 0; j < m_models[0].length; ++j) {
                    for (int k = 0; k < m_models[0][j].length; ++k) {
                        predictions[j][k] += model(i)[j][k] * m_timesChosen[i];
                    }
                }
            }
        }
        for (int j = 0; j < m_models[0].length; ++j) {
            int change = add ? 1 : -1;
            for (int k = 0; k < m_models[0][j].length; ++k) {
                predictions[j][k] += change * model(index_to_change)[j][k];
                predictions[j][k] /= (m_numChosen + change);
            }
        }
        return predictions;
    }

    private double evaluatePredictions(Instances instances,
            double[][] temp_predictions, int metric) throws Exception {

        Evaluation eval = new Evaluation(instances);
        for (int i = 0; i < instances.numInstances(); ++i) {
            eval.evaluateModelOnceAndRecordPrediction(temp_predictions[i],
                    instances.instance(i));
        }
        return BESHelper.getMetric(eval, metric);
    }

    public double[] getIndividualPerformance(Instances instances, int metric)
            throws Exception {

        double[] performance = new double[m_bagSize];
        for (int i = 0; i < m_bagSize; ++i) {
            performance[i] = evaluatePredictions(instances, model(i), metric);
        }
        return performance;
    }

    public String getRevision() {
        return RevisionUtils.extract("$Revision: 1.2 $");
    }
}
