/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    BESnnls.java
 *    Copyright (C) 1999-2012 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.meta.baggingEnsembleSelection;
import java.io.Serializable;
import java.lang.reflect.Method;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;
import weka.classifiers.*;
import weka.core.*;
import weka.core.matrix.FlexibleDecimalFormat;
import weka.core.matrix.Maths;
import weka.core.matrix.Matrix;
import weka.filters.Filter;

/**
 * <!-- globalinfo-start --> Class for using non-negative least-squares for
 * prediction.
 * <p/>
 * <!-- globalinfo-end -->
 *
 * <!-- options-start --> Valid options are:
 * <p/>
 *
 * <pre> -D
 *  Produce debugging output.
 *  (default no debugging output)</pre>
 *
 * <pre> -O
 *  Force the regression coefficients sum to 1.
 * </pre>
 *
 * <!-- options-end -->
 *
 * @author Quan Sun (qs12@cs.waikato.ac.nz)
 * @author The Mathworks and NIST
 * @author Yong Wang
 * @version $Revision: 8034 $
 */
public class BESnnls extends AbstractClassifier implements OptionHandler,
        TechnicalInformationHandler {

    /**
     * for serialization
     */
    static final long serialVersionUID = 5655510626304705059L;
    /**
     * Array for storing coefficients of NNLS regression.
     */
    private NNLSDoubleVector m_Coefficients;
    /**
     * Array for set P.
     */
    private NNLSIntVector m_Set_P;
    /**
     * Force coefficients sum to 1.0?
     */
    private boolean m_ForceCoefficientsSumToOne = false;
    /**
     * A copy of the dataset information.
     */
    private Instances m_Header;
    /**
     * Missing value replacement filter.
     */
    private weka.filters.unsupervised.attribute.ReplaceMissingValues m_ReplaceMissingValues;

    /**
     * Returns an instance of a TechnicalInformation object, containing detailed
     * information about the technical background of this class, e.g., paper
     * reference or book this class is based on.
     *
     * @return the technical information about this class
     */
    @Override
    public TechnicalInformation getTechnicalInformation() {
        TechnicalInformation result;
        result = new TechnicalInformation(TechnicalInformation.Type.ARTICLE);
        result.setValue(TechnicalInformation.Field.AUTHOR, "Lawson, C. L. and Hanson, B. J");
        result.setValue(TechnicalInformation.Field.YEAR, "1974");
        result.setValue(TechnicalInformation.Field.PUBLISHER, "Prentice-Hall");
        result.setValue(TechnicalInformation.Field.BOOKTITLE, "Solving Least Squares Problems");
        return result;
    }

    /**
     * Sets the method used to force coefficients sum to 1.0 regression.
     *
     * @param o if the sum of coefficients is forced to equal 1.0
     */
    public void setForceCoefficientsSumToOne(boolean o) {
        m_ForceCoefficientsSumToOne = o;
    }

    /**
     * Returns whether the sum of coefficients is forced to equal 1.0.
     *
     * @return true if the sum of coefficients is forced to equal 1.0
     */
    public boolean getForceCoefficientsSumToOne() {
        return m_ForceCoefficientsSumToOne;
    }

    /**
     * Builds a NNLS regression model for the given data.
     *
     * @param data the training data to be used for generating the linear
     * regression function
     * @throws Exception if the classifier could not be built successfully
     */
    @Override
    public void buildClassifier(Instances data) {

        Instances workingData = null;
        try {
            m_ReplaceMissingValues = new weka.filters.unsupervised.attribute.ReplaceMissingValues();
            m_ReplaceMissingValues.setInputFormat(data);
            workingData = Filter.useFilter(data, m_ReplaceMissingValues);
        } catch (Exception e) {
            System.err.println(e.toString());
        }

        m_Header = new Instances(data, 0);

        int numDataPoints = workingData.numInstances();
        int numIndepVariables = workingData.numAttributes() - 1;

        NNLSMatrix A = new NNLSMatrix(numDataPoints, numIndepVariables);
        NNLSMatrix b = new NNLSMatrix(numDataPoints, 1);

        for (int row = 0; row < numDataPoints; row++) {
            Instance inst = workingData.instance(row);

            int count = 0;

            for (int col = 0; col < numIndepVariables; col++) {
                if (m_Header.classIndex() != col) {
                    A.set(row, count, inst.value(col));
                }
                count++;
            }

            b.set(row, 0, inst.classValue());
        }

        m_Set_P = new NNLSIntVector(numIndepVariables);

        for (int col = 0; col < numIndepVariables; col++) {
            m_Set_P.set(col, col);
        }

        if (m_ForceCoefficientsSumToOne) {
            m_Coefficients = A.nnlse1(b, m_Set_P);
        } else {
            m_Coefficients = A.nnls(b, m_Set_P);
        }
    }

    /**
     * Returns the actual attribute index.
     *
     * @param index Index.
     * @return the actual attribute index
     */
    private int mapIndex(int index) {
        if (index >= m_Header.classIndex()) {
            return index + 1;
        }
        return index;
    }

    /**
     * Returns the coefficients for this linear model.
     *
     * @param attIndex Index of an attribute.
     * @return the coefficients for this linear model
     */
    public double getCoefficient(int attIndex) {
        for (int i = 0; i < m_Coefficients.size(); i++) {
            double c = m_Coefficients.get(i);
            int oriIndex = mapIndex(m_Set_P.get(i));

            if (attIndex == oriIndex) {
                return c;
            }
        }
        return Double.NaN;
    }

    /**
     * Classifies the given instance using the linear regression function.
     *
     * @param instance the test instance
     * @return the classification
     * @throws Exception if classification can't be done successfully
     */
    @Override
    public double classifyInstance(Instance datapoint) throws Exception {
        Instance newInst = (Instance) datapoint.copy();
        m_ReplaceMissingValues.input(newInst);
        newInst = m_ReplaceMissingValues.output();

        double pred = 0;
        for (int i = 0; i < m_Coefficients.size(); i++) {
            double c = m_Coefficients.get(i);
            int oriIndex = mapIndex(m_Set_P.get(i));

            if (oriIndex != -1) {
                pred += c * newInst.value(oriIndex);
            }
        }
        return pred;
    }

    /**
     * Returns a string describing this classifier
     *
     * @return a description of the classifier suitable for displaying in the
     * explorer/experimenter gui
     */
    public String globalInfo() {
        return "Non-negative least-squares (NNLS), introduced by Lawson & Hanson (1974) "
                + ", the coefficients contain no negative elements. ";
    }

    /**
     * Returns default capabilities of the classifier.
     *
     * @return the capabilities of this classifier
     */
    @Override
    public Capabilities getCapabilities() {
        Capabilities result = super.getCapabilities();
        result.disableAll();
        // attributes        
        result.enable(Capabilities.Capability.NUMERIC_ATTRIBUTES);
        result.enable(Capabilities.Capability.MISSING_VALUES);
        // class
        result.enable(Capabilities.Capability.NUMERIC_CLASS);
        return result;
    }

    /**
     * Returns an enumeration describing the available options.
     *
     * @return an enumeration of all the available options.
     */
    @Override
    public Enumeration listOptions() {
        Vector newVector = new Vector();

        newVector.addElement(
                new Option(
                "\tProduce debugging output.\n"
                + "\t(default no debugging output)",
                "D", 0, "-D"));

        newVector.addElement(
                new Option(
                "\t The sum of coefficients equals 1. \n",
                "O", 0, "-O"));
        newVector.addElement(
                new Option(
                "\t Add the intercept term. \n",
                "Z", 0, "-Z"));
        return newVector.elements();
    }

    /**
     * Gets the current settings of the classifier.
     *
     * @return an array of strings suitable for passing to setOptions
     */
    @Override
    public String[] getOptions() {
        Vector<String> result;
        result = new Vector<String>();
        if (getDebug()) {
            result.add("-D");
        }

        if (getForceCoefficientsSumToOne()) {
            result.add("-O");
        }
        return result.toArray(new String[result.size()]);
    }

    /**
     * Parses a given list of options.
     * <p/>
     *
     * <!-- options-start --> Valid options are:
     * <p/>
     *
     * <pre> -D
     *  Produce debugging output.
     *  (default no debugging output)</pre>
     *
     * <pre> -O
     *  Force the sum of coefficients to 1.0.
     * </pre>
     *
     * <!-- options-end -->
     *
     * @param options the list of options as an array of strings
     * @throws Exception if an option is not supported
     */
    @Override
    public void setOptions(String[] options) throws Exception {
        setDebug(Utils.getFlag('D', options));
        setForceCoefficientsSumToOne(!Utils.getFlag('O', options));      
        super.setOptions(options);
    }

    /**
     * Returns the revision string.
     *
     * @return	the revision
     */
    public String getRevision() {
        return RevisionUtils.extract("$Revision: 8034 $");
    }

    /**
     * @author The Mathworks and NIST
     * @author Yong Wang
     */
    private static class NNLSIntVector
            implements Cloneable, Serializable {

        /**
         * Array for internal storage of elements.
         */
        int[] V;
        /**
         * size of the vector
         */
        private int sizeOfVector;


        /* ------------------------
         Constructors
         * ------------------------ */
        /**
         * Constructs a null vector.
         */
        public NNLSIntVector() {
            V = new int[0];
            setSize(0);
        }

        /**
         * Constructs an n-vector of zeros.
         *
         * @param n Length.
         */
        public NNLSIntVector(int n) {
            V = new int[n];
            setSize(n);
        }

        /**
         * Constructs an n-vector of a constant
         *
         * @param n Length.
         */
        public NNLSIntVector(int n, int s) {
            this(n);
            set(s);
        }

        /**
         * Constructs a vector given an int array
         *
         * @param v the int array
         */
        public NNLSIntVector(int v[]) {
            if (v == null) {
                V = new int[0];
                setSize(0);
            } else {
                V = new int[v.length];
                setSize(v.length);
                set(0, size() - 1, v, 0);
            }
        }

        /* ------------------------
         Public Methods
         * ------------------------ */
        /**
         * Gets the size of the vector.
         *
         * @return Size.
         */
        public int size() {
            return sizeOfVector;
        }

        /**
         * Sets the size of the vector. The provided size can't be greater than
         * the capacity of the vector.
         *
         * @param size the new Size.
         */
        public void setSize(int size) {
            if (size > capacity()) {
                throw new IllegalArgumentException("insufficient capacity");
            }
            sizeOfVector = size;
        }

        /**
         * Sets the value of an element.
         *
         * @param s the value for the element
         */
        public void set(int s) {
            for (int i = 0; i < size(); i++) {
                set(i, s);
            }
        }

        /**
         * Sets the values of elements from an int array.
         *
         * @param i0 the index of the first element
         * @param i1 the index of the last element
         * @param v the int array that stores the values
         * @param j0 the index of the first element in the int array
         */
        public void set(int i0, int i1, int[] v, int j0) {
            for (int i = i0; i <= i1; i++) {
                set(i, v[j0 + i - i0]);
            }
        }

        /**
         * Sets the values of elements from another IntVector.
         *
         * @param i0 the index of the first element
         * @param i1 the index of the last element
         * @param v the IntVector that stores the values
         * @param j0 the index of the first element in the IntVector
         */
        public void set(int i0, int i1, NNLSIntVector v, int j0) {
            for (int i = i0; i <= i1; i++) {
                set(i, v.get(j0 + i - i0));
            }
        }

        /**
         * Sets the values of elements from another IntVector.
         *
         * @param v the IntVector that stores the values
         */
        public void set(NNLSIntVector v) {
            set(0, v.size() - 1, v, 0);
        }

        /**
         * Generates an IntVector that stores all integers inclusively between
         * two integers.
         *
         * @param i0 the first integer
         * @param i1 the second integer
         */
        public static NNLSIntVector seq(int i0, int i1) {
            if (i1 < i0) {
                throw new IllegalArgumentException("i1 < i0 ");
            }
            NNLSIntVector v = new NNLSIntVector(i1 - i0 + 1);
            for (int i = 0; i < i1 - i0 + 1; i++) {
                v.set(i, i + i0);
            }
            return v;
        }

        /**
         * Access the internal one-dimensional array.
         *
         * @return Pointer to the one-dimensional array of vector elements.
         */
        public int[] getArray() {
            return V;
        }

        /**
         * Sets the internal one-dimensional array.
         *
         * @param a Pointer to the one-dimensional array of vector elements.
         */
        protected void setArray(int[] a) {
            V = a;
        }

        /**
         * Sorts the elements in place
         */
        public void sort() {
            Arrays.sort(V, 0, size());
        }

        /**
         * Returns a copy of the internal one-dimensional array.
         *
         * @return One-dimensional array copy of vector elements.
         */
        public int[] getArrayCopy() {
            int[] b = new int[size()];
            for (int i = 0; i <= size() - 1; i++) {
                b[i] = V[i];
            }
            return b;
        }

        /**
         * Returns the capacity of the vector
         */
        public int capacity() {
            return V.length;
        }

        /**
         * Sets the capacity of the vector
         *
         * @param capacity the new capacity of the vector
         */
        public void setCapacity(int capacity) {
            if (capacity == capacity()) {
                return;
            }
            int[] old_V = V;
            int m = Math.min(capacity, size());
            V = new int[capacity];
            setSize(capacity);
            set(0, m - 1, old_V, 0);
        }

        /**
         * Sets a single element.
         *
         * @param i the index of the element
         * @param s the new value
         */
        public void set(int i, int s) {
            V[i] = s;
        }

        /**
         * Gets the value of an element.
         *
         * @param i the index of the element
         * @return the value of the element
         */
        public int get(int i) {
            return V[i];
        }

        /**
         * Makes a deep copy of the vector
         */
        public NNLSIntVector copy() {
            return (NNLSIntVector) clone();
        }

        /**
         * Clones the IntVector object.
         */
        public Object clone() {
            NNLSIntVector u = new NNLSIntVector(size());
            for (int i = 0; i < size(); i++) {
                u.V[i] = V[i];
            }
            return u;
        }

        /**
         * Returns a subvector.
         *
         * @param i0 the index of the first element
         * @param i1 the index of the last element
         * @return the subvector
         */
        public NNLSIntVector subvector(int i0, int i1) {
            NNLSIntVector v = new NNLSIntVector(i1 - i0 + 1);
            v.set(0, i1 - i0, this, i0);
            return v;
        }

        /**
         * Returns a subvector as indexed by an IntVector.
         *
         * @param index the index
         * @return the subvector
         */
        public NNLSIntVector subvector(NNLSIntVector index) {
            NNLSIntVector v = new NNLSIntVector(index.size());
            for (int i = 0; i < index.size(); i++) {
                v.V[i] = V[index.V[i]];
            }
            return v;
        }

        /**
         * Swaps the values stored at i and j
         *
         * @param i the index i
         * @param j the index j
         */
        public void swap(int i, int j) {
            if (i == j) {
                return;
            }
            int t = get(i);
            set(i, get(j));
            set(j, t);
        }

        /**
         * Shifts an element to another position. Elements between them are
         * shifted one position left.
         *
         * @param i the index of the element
         * @param j the index of the new position
         */
        public void shift(int i, int j) {
            if (i == j) {
                return;
            }
            if (i < j) {
                int t = V[i];
                for (int k = i; k <= j - 1; k++) {
                    V[k] = V[k + 1];
                }
                V[j] = t;
            } else {
                shift(j, i);
            }
        }

        /**
         * Shifts an element to the end of the vector. Elements between them are
         * shifted one position left.
         *
         * @param j the index of the element
         */
        public void shiftToEnd(int j) {
            shift(j, size() - 1);
        }

        /**
         * Returns true if the vector is empty
         */
        public boolean isEmpty() {
            if (size() == 0) {
                return true;
            }
            return false;
        }
    }

    /**
     * @author The Mathworks and NIST
     * @author Yong Wang
     */
    private static class NNLSDoubleVector
            implements Cloneable, Serializable {

        double[] V; // array for internal storage of elements.
        private int sizeOfVector;      // size of the vector

        /* ------------------------
         Constructors
         * ------------------------ */
        /**
         * Constructs a null vector.
         */
        public NNLSDoubleVector() {
            this(0);
        }

        /**
         * Constructs an n-vector of zeros.
         *
         * @param n length.
         */
        public NNLSDoubleVector(int n) {
            V = new double[n];
            setSize(n);
        }

        /**
         * Constructs a constant n-vector.
         *
         * @param n length.
         * @param s the scalar value used to fill the vector
         */
        public NNLSDoubleVector(int n, double s) {
            this(n);
            set(s);
        }

        /**
         * Constructs a vector directly from a double array
         *
         * @param v the array
         */
        public NNLSDoubleVector(double v[]) {
            if (v == null) {
                V = new double[0];
                setSize(0);
            } else {
                V = v;
                setSize(v.length);
            }
        }

        /* ------------------------
         *  Public Methods
         * ------------------------ */
        /**
         * Set a single element.
         *
         * @param i Index.
         * @param s a[i].
         */
        public void set(int i, double s) {

            V[i] = s;
        }

        /**
         * Set all elements to a value
         *
         * @param s the value
         */
        public void set(double s) {
            set(0, size() - 1, s);
        }

        /**
         * Set some elements to a value
         *
         * @param i0 the index of the first element
         * @param i1 the index of the second element
         * @param s the value
         */
        public void set(int i0, int i1, double s) {

            for (int i = i0; i <= i1; i++) {
                V[i] = s;
            }
        }

        /**
         * Set some elements using a 2-D array
         *
         * @param i0 the index of the first element
         * @param i1 the index of the second element
         * @param j0 the index of the starting element in the 2-D array
         * @param v the values
         */
        public void set(int i0, int i1, double[] v, int j0) {
            for (int i = i0; i <= i1; i++) {
                V[i] = v[j0 + i - i0];
            }
        }

        /**
         * Set the elements using a DoubleVector
         *
         * @param v the DoubleVector
         */
        public void set(NNLSDoubleVector v) {
            set(0, v.size() - 1, v, 0);
        }

        /**
         * Set some elements using a DoubleVector.
         *
         * @param i0 the index of the first element
         * @param i1 the index of the second element
         * @param v the DoubleVector
         * @param j0 the index of the starting element in the DoubleVector
         */
        public void set(int i0, int i1, NNLSDoubleVector v, int j0) {
            for (int i = i0; i <= i1; i++) {
                V[i] = v.V[j0 + i - i0];
            }
        }

        /**
         * Access the internal one-dimensional array.
         *
         * @return Pointer to the one-dimensional array of vector elements.
         */
        public double[] getArray() {
            return V;
        }

        void setArray(double[] a) {
            V = a;
        }

        /**
         * Returns a copy of the DoubleVector usng a double array.
         *
         * @return the one-dimensional array.
         */
        public double[] getArrayCopy() {
            double v[] = new double[size()];

            for (int i = 0; i < size(); i++) {
                v[i] = V[i];
            }

            return v;
        }

        /**
         * Sorts the array in place
         */
        public void sort() {
            Arrays.sort(V, 0, size());
        }

        /**
         * Sorts the array in place with index returned
         */
        public NNLSIntVector sortWithIndex() {
            NNLSIntVector index = NNLSIntVector.seq(0, size() - 1);
            sortWithIndex(0, size() - 1, index);
            return index;
        }

        /**
         * Sorts the array in place with index changed
         *
         * @param xi first index
         * @param xj last index
         * @param index array that stores all indices
         */
        public void sortWithIndex(int xi, int xj, NNLSIntVector index) {
            if (xi < xj) {
                double x, f, k;
                int xm = (int) (xi + xj) / 2; // median index
                x = Math.min(V[xi], // median of three
                        Math.max(V[xm], V[xj]));
                int i = xi;
                int j = xj;
                while (i < j) {
                    while (V[i] < x && i < xj) {
                        i++;
                    }
                    while (V[j] > x && j > xi) {
                        j--;
                    }
                    if (i <= j) {
                        swap(i, j);
                        index.swap(i, j);
                        i++;
                        j--;
                    }
                }
                sortWithIndex(xi, j, index);
                sortWithIndex(i, xj, index);
            }
        }

        /**
         * Gets the size of the vector.
         *
         * @return the size
         */
        public int size() {
            return sizeOfVector;
        }

        /**
         * Sets the size of the vector
         *
         * @param m the size
         */
        public void setSize(int m) {
            if (m > capacity()) {
                throw new IllegalArgumentException("insufficient capacity");
            }
            sizeOfVector = m;
        }

        /**
         * Gets the capacity of the vector.
         *
         * @return the capacity.
         */
        public int capacity() {
            if (V == null) {
                return 0;
            }
            return V.length;
        }

        /**
         * Sets the capacity of the vector
         *
         * @param n the capacity.
         */
        public void setCapacity(int n) {
            if (n == capacity()) {
                return;
            }
            double[] oldV = V;
            int m = Math.min(n, size());
            V = new double[n];
            setSize(m);
            set(0, m - 1, oldV, 0);
        }

        /**
         * Gets a single element.
         *
         * @param i Index.
         * @return the value of the i-th element
         */
        public double get(int i) {
            return V[i];
        }

        /**
         * Adds a value to an element
         *
         * @param i the index of the element
         * @param s the value
         */
        public void setPlus(int i, double s) {
            V[i] += s;
        }

        /**
         * Multiplies a value to an element
         *
         * @param i the index of the element
         * @param s the value
         */
        public void setTimes(int i, double s) {
            V[i] *= s;
        }

        /**
         * Adds an element into the vector
         *
         * @param x the value of the new element
         */
        public void addElement(double x) {
            if (capacity() == 0) {
                setCapacity(10);
            }
            if (size() == capacity()) {
                setCapacity(2 * capacity());
            }
            V[size()] = x;
            setSize(size() + 1);
        }

        /**
         * Returns the squared vector
         */
        public NNLSDoubleVector square() {
            NNLSDoubleVector v = new NNLSDoubleVector(size());
            for (int i = 0; i < size(); i++) {
                v.V[i] = V[i] * V[i];
            }
            return v;
        }

        /**
         * Returns the square-root of all the elements in the vector
         */
        public NNLSDoubleVector sqrt() {
            NNLSDoubleVector v = new NNLSDoubleVector(size());
            for (int i = 0; i < size(); i++) {
                v.V[i] = Math.sqrt(V[i]);
            }
            return v;
        }

        /**
         * Makes a deep copy of the vector
         */
        public NNLSDoubleVector copy() {
            return (NNLSDoubleVector) clone();
        }

        /**
         * Clones the DoubleVector object.
         */
        public Object clone() {
            int n = size();
            NNLSDoubleVector u = new NNLSDoubleVector(n);
            for (int i = 0; i < n; i++) {
                u.V[i] = V[i];
            }
            return u;
        }

        /**
         * Returns the inner product of two DoubleVectors
         *
         * @param v the second DoubleVector
         * @return the product
         */
        public double innerProduct(NNLSDoubleVector v) {
            if (size() != v.size()) {
                throw new IllegalArgumentException("sizes unmatch");
            }
            double p = 0;
            for (int i = 0; i < size(); i++) {
                p += V[i] * v.V[i];
            }
            return p;
        }

        /**
         * Returns the signs of all elements in terms of -1, 0 and +1.
         */
        public NNLSDoubleVector sign() {
            NNLSDoubleVector s = new NNLSDoubleVector(size());
            for (int i = 0; i < size(); i++) {
                if (V[i] > 0) {
                    s.V[i] = 1;
                } else if (V[i] < 0) {
                    s.V[i] = -1;
                } else {
                    s.V[i] = 0;
                }
            }
            return s;
        }

        /**
         * Returns the sum of all elements in the vector.
         */
        public double sum() {
            double s = 0;
            for (int i = 0; i < size(); i++) {
                s += V[i];
            }
            return s;
        }

        /**
         * Returns the squared sum of all elements in the vector.
         */
        public double sum2() {
            double s2 = 0;
            for (int i = 0; i < size(); i++) {
                s2 += V[i] * V[i];
            }
            return s2;
        }

        /**
         * Returns the L1-norm of the vector
         */
        public double norm1() {
            double s = 0;
            for (int i = 0; i < size(); i++) {
                s += Math.abs(V[i]);
            }
            return s;
        }

        /**
         * Returns the L2-norm of the vector
         */
        public double norm2() {
            return Math.sqrt(sum2());
        }

        /**
         * Returns ||u-v||^2
         *
         * @param v the second vector
         */
        public double sum2(NNLSDoubleVector v) {
            return minus(v).sum2();
        }

        /**
         * Returns a subvector.
         *
         * @param i0 the index of the first element
         * @param i1 the index of the last element
         * @return v[i0:i1]
         */
        public NNLSDoubleVector subvector(int i0, int i1) {
            NNLSDoubleVector v = new NNLSDoubleVector(i1 - i0 + 1);
            v.set(0, i1 - i0, this, i0);
            return v;
        }

        /**
         * Returns a subvector.
         *
         * @param index stores the indices of the needed elements
         * @return v[index]
         */
        public NNLSDoubleVector subvector(NNLSIntVector index) {
            NNLSDoubleVector v = new NNLSDoubleVector(index.size());
            for (int i = 0; i < index.size(); i++) {
                v.V[i] = V[index.V[i]];
            }
            return v;
        }

        /**
         * Returns a vector from the pivoting indices. Elements not indexed are
         * set to zero.
         *
         * @param index stores the pivoting indices
         * @param length the total number of the potential elements
         * @return the subvector
         */
        public NNLSDoubleVector unpivoting(NNLSIntVector index, int length) {
            if (index.size() > length) {
                throw new IllegalArgumentException("index.size() > length ");
            }
            NNLSDoubleVector u = new NNLSDoubleVector(length);
            for (int i = 0; i < index.size(); i++) {
                u.V[index.V[i]] = V[i];
            }
            return u;
        }

        /**
         * Adds a value to all the elements
         *
         * @param x the value
         */
        public NNLSDoubleVector plus(double x) {
            return copy().plusEquals(x);
        }

        /**
         * Adds a value to all the elements in place
         *
         * @param x the value
         */
        public NNLSDoubleVector plusEquals(double x) {
            for (int i = 0; i < size(); i++) {
                V[i] += x;
            }
            return this;
        }

        /**
         * Adds another vector element by element
         *
         * @param v the second vector
         */
        public NNLSDoubleVector plus(NNLSDoubleVector v) {
            return copy().plusEquals(v);
        }

        /**
         * Adds another vector in place element by element
         *
         * @param v the second vector
         */
        public NNLSDoubleVector plusEquals(NNLSDoubleVector v) {
            for (int i = 0; i < size(); i++) {
                V[i] += v.V[i];
            }
            return this;
        }

        /**
         * Subtracts a value
         *
         * @param x the value
         */
        public NNLSDoubleVector minus(double x) {
            return plus(-x);
        }

        /**
         * Subtracts a value in place
         *
         * @param x the value
         */
        public NNLSDoubleVector minusEquals(double x) {
            plusEquals(-x);
            return this;
        }

        /**
         * Subtracts another DoubleVector element by element
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector minus(NNLSDoubleVector v) {
            return copy().minusEquals(v);
        }

        /**
         * Subtracts another DoubleVector element by element in place
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector minusEquals(NNLSDoubleVector v) {
            for (int i = 0; i < size(); i++) {
                V[i] -= v.V[i];
            }
            return this;
        }

        /**
         * Multiplies a scalar
         *
         * @param s scalar
         * @return s * v
         */
        public NNLSDoubleVector times(double s) {
            return copy().timesEquals(s);
        }

        /**
         * Multiply a vector by a scalar in place, u = s * u
         *
         * @param s scalar
         * @return replace u by s * u
         */
        public NNLSDoubleVector timesEquals(double s) {
            for (int i = 0; i < size(); i++) {
                V[i] *= s;
            }
            return this;
        }

        /**
         * Multiplies another DoubleVector element by element
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector times(NNLSDoubleVector v) {
            return copy().timesEquals(v);

        }

        /**
         * Multiplies another DoubleVector element by element in place
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector timesEquals(NNLSDoubleVector v) {
            for (int i = 0; i < size(); i++) {
                V[i] *= v.V[i];
            }
            return this;
        }

        /**
         * Divided by another DoubleVector element by element
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector dividedBy(NNLSDoubleVector v) {
            return copy().dividedByEquals(v);
        }

        /**
         * Divided by another DoubleVector element by element in place
         *
         * @param v the second DoubleVector
         */
        public NNLSDoubleVector dividedByEquals(NNLSDoubleVector v) {
            for (int i = 0; i < size(); i++) {
                V[i] /= v.V[i];
            }
            return this;
        }

        /**
         * Checks if it is an empty vector
         */
        public boolean isEmpty() {
            if (size() == 0) {
                return true;
            }
            return false;
        }

        /**
         * Returns a vector that stores the cumulated values of the original
         * vector
         */
        public NNLSDoubleVector cumulate() {
            return copy().cumulateInPlace();
        }

        /**
         * Cumulates the original vector in place
         */
        public NNLSDoubleVector cumulateInPlace() {
            for (int i = 1; i < size(); i++) {
                V[i] += V[i - 1];
            }
            return this;
        }

        /**
         * Returns the index of the maximum. <p> If multiple maximums exist, the
         * index of the first is returned.
         */
        public int indexOfMax() {
            int index = 0;
            double ma = V[0];

            for (int i = 1; i < size(); i++) {
                if (ma < V[i]) {
                    ma = V[i];
                    index = i;
                }
            }
            return index;
        }

        /**
         * Returns true if vector not sorted
         */
        public boolean unsorted() {
            if (size() < 2) {
                return false;
            }
            for (int i = 1; i < size(); i++) {
                if (V[i - 1] > V[i]) {
                    return true;
                }
            }
            return false;
        }

        /**
         * Combine two vectors together
         *
         * @param v the second vector
         */
        public NNLSDoubleVector cat(NNLSDoubleVector v) {
            NNLSDoubleVector w = new NNLSDoubleVector(size() + v.size());
            w.set(0, size() - 1, this, 0);
            w.set(size(), size() + v.size() - 1, v, 0);
            return w;
        }

        /**
         * Swaps the values stored at i and j
         *
         * @param i the index i
         * @param j the index j
         */
        public void swap(int i, int j) {
            if (i == j) {
                return;
            }
            double t = V[i];
            V[i] = V[j];
            V[j] = t;
        }

        /**
         * Returns the maximum value of all elements
         */
        public double max() {
            if (size() < 1) {
                throw new IllegalArgumentException("zero size");
            }
            double ma = V[0];
            if (size() < 2) {
                return ma;
            }
            for (int i = 1; i < size(); i++) {
                if (V[i] > ma) {
                    ma = V[i];
                }
            }
            return ma;
        }

        /**
         * Applies a method to the vector
         *
         * @param className the class name
         * @param method the method
         */
        public NNLSDoubleVector map(String className, String method) {
            try {
                Class<?> c = Class.forName(className);
                Class[] cs = new Class[1];
                cs[ 0] = Double.TYPE;
                Method m = c.getMethod(method, cs);

                NNLSDoubleVector w = new NNLSDoubleVector(size());
                Object[] obj = new Object[1];
                for (int i = 0; i < size(); i++) {
                    obj[0] = new Double(V[i]);
                    w.set(i, Double.parseDouble(m.invoke(null, obj).toString()));
                }
                return w;
            } catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
            return null;
        }

        /**
         * Returns the reverse vector
         */
        public NNLSDoubleVector rev() {
            int n = size();
            NNLSDoubleVector w = new NNLSDoubleVector(n);
            for (int i = 0; i < n; i++) {
                w.V[i] = V[n - i - 1];
            }
            return w;
        }

        /**
         * Returns a random vector of uniform distribution
         *
         * @param n the size of the vector
         */
        public static NNLSDoubleVector random(int n) {
            NNLSDoubleVector v = new NNLSDoubleVector(n);
            for (int i = 0; i < n; i++) {
                v.V[i] = Math.random();
            }
            return v;
        }
    }

    /**
     * @author The Mathworks and NIST
     * @author Yong Wang
     */
    public static class NNLSMatrix
            extends Matrix {

        /* ------------------------
         Constructors
         * ------------------------ */
        /**
         * Construct an m-by-n PACE matrix of zeros.
         *
         * @param m Number of rows.
         * @param n Number of colums.
         */
        public NNLSMatrix(int m, int n) {
            super(m, n);
        }

        /**
         * Construct an m-by-n constant PACE matrix.
         *
         * @param m Number of rows.
         * @param n Number of colums.
         * @param s Fill the matrix with this scalar value.
         */
        public NNLSMatrix(int m, int n, double s) {
            super(m, n, s);
        }

        /**
         * Construct a PACE matrix from a 2-D array.
         *
         * @param A Two-dimensional array of doubles.
         * @throws IllegalArgumentException All rows must have the same length
         */
        public NNLSMatrix(double[][] A) {
            super(A);
        }

        /**
         * Construct a PACE matrix quickly without checking arguments.
         *
         * @param A Two-dimensional array of doubles.
         * @param m Number of rows.
         * @param n Number of colums.
         */
        public NNLSMatrix(double[][] A, int m, int n) {
            super(A, m, n);
        }

        /**
         * Construct a PaceMatrix from a one-dimensional packed array
         *
         * @param vals One-dimensional array of doubles, packed by columns (ala
         * Fortran).
         * @param m Number of rows.
         * @throws IllegalArgumentException Array length must be a multiple of
         * m.
         */
        public NNLSMatrix(double vals[], int m) {
            super(vals, m);
        }

        /**
         * Construct a PaceMatrix from a Matrix
         *
         * @param X Matrix
         */
        public NNLSMatrix(Matrix X) {
            super(X.getRowDimension(), X.getColumnDimension());
            A = X.getArray();
        }

        /* ------------------------
         Public Methods
         * ------------------------ */
        /**
         * Set the row dimenion of the matrix
         *
         * @param rowDimension the row dimension
         */
        public void setRowDimension(int rowDimension) {
            m = rowDimension;
        }

        /**
         * Set the column dimenion of the matrix
         *
         * @param columnDimension the column dimension
         */
        public void setColumnDimension(int columnDimension) {
            n = columnDimension;
        }

        /**
         * Clone the PaceMatrix object.
         *
         * @return the clone
         */
        public Object clone() {
            NNLSMatrix X = new NNLSMatrix(m, n);
            double[][] C = X.getArray();
            for (int i = 0; i < m; i++) {
                for (int j = 0; j < n; j++) {
                    C[i][j] = A[i][j];
                }
            }
            return (Object) X;
        }

        /**
         * Add a value to an element and reset the element
         *
         * @param i the row number of the element
         * @param j the column number of the element
         * @param s the double value to be added with
         */
        public void setPlus(int i, int j, double s) {
            A[i][j] += s;
        }

        /**
         * Returns the maximum absolute value of all elements
         *
         * @return the maximum value
         */
        public double maxAbs() {
            double ma = Math.abs(A[0][0]);
            for (int j = 0; j < n; j++) {
                for (int i = 0; i < m; i++) {
                    ma = Math.max(ma, Math.abs(A[i][j]));
                }
            }
            return ma;
        }

        /**
         * Returns the maximum absolute value of some elements of a column, that
         * is, the elements of A[i0:i1][j].
         *
         * @param i0 the index of the first element of the column
         * @param i1 the index of the last element of the column
         * @param j the index of the column
         * @return the maximum value
         */
        public double maxAbs(int i0, int i1, int j) {
            double m = Math.abs(A[i0][j]);
            for (int i = i0 + 1; i <= i1; i++) {
                m = Math.max(m, Math.abs(A[i][j]));
            }
            return m;
        }

        /**
         * Returns the minimum absolute value of some elements of a column, that
         * is, the elements of A[i0:i1][j].
         *
         * @param i0 the index of the first element of the column
         * @param i1 the index of the last element of the column
         * @param column the index of the column
         * @return the minimum value
         */
        public double minAbs(int i0, int i1, int column) {
            double m = Math.abs(A[i0][column]);
            for (int i = i0 + 1; i <= i1; i++) {
                m = Math.min(m, Math.abs(A[i][column]));
            }
            return m;
        }

        /**
         * Check if the matrix is empty
         *
         * @return true if the matrix is empty
         */
        public boolean isEmpty() {
            if (m == 0 || n == 0) {
                return true;
            }
            if (A == null) {
                return true;
            }
            return false;
        }

        /**
         * Return a DoubleVector that stores a column of the matrix
         *
         * @param j the index of the column
         * @return the column
         */
        public NNLSDoubleVector getColumn(int j) {
            NNLSDoubleVector v = new NNLSDoubleVector(m);
            double[] a = v.getArray();
            for (int i = 0; i < m; i++) {
                a[i] = A[i][j];
            }
            return v;
        }

        /**
         * Return a DoubleVector that stores some elements of a column of the
         * matrix
         *
         * @param i0 the index of the first element of the column
         * @param i1 the index of the last element of the column
         * @param j the index of the column
         * @return the DoubleVector
         */
        public NNLSDoubleVector getColumn(int i0, int i1, int j) {
            NNLSDoubleVector v = new NNLSDoubleVector(i1 - i0 + 1);
            double[] a = v.getArray();
            int count = 0;
            for (int i = i0; i <= i1; i++) {
                a[count] = A[i][j];
                count++;
            }
            return v;
        }

        /**
         * Multiplication between a row (or part of a row) of the first matrix
         * and a column (or part or a column) of the second matrix.
         *
         * @param i the index of the row in the first matrix
         * @param j0 the index of the first column in the first matrix
         * @param j1 the index of the last column in the first matrix
         * @param B the second matrix
         * @param l the index of the column in the second matrix
         * @return the result of the multiplication
         */
        public double times(int i, int j0, int j1, NNLSMatrix B, int l) {
            double s = 0.0;
            for (int j = j0; j <= j1; j++) {
                s += A[i][j] * B.A[j][l];
            }
            return s;
        }

        /**
         * Decimal format for converting a matrix into a string
         *
         * @return the default decimal format
         */
        protected DecimalFormat[] format() {
            return format(0, m - 1, 0, n - 1, 7, false);
        }

        /**
         * Decimal format for converting a matrix into a string
         *
         * @param digits the number of digits
         * @return the decimal format
         */
        protected DecimalFormat[] format(int digits) {
            return format(0, m - 1, 0, n - 1, digits, false);
        }

        /**
         * Decimal format for converting a matrix into a string
         *
         * @param digits the number of digits
         * @param trailing
         * @return the decimal format
         */
        protected DecimalFormat[] format(int digits, boolean trailing) {
            return format(0, m - 1, 0, n - 1, digits, trailing);
        }

        /**
         * Decimal format for converting a matrix into a string
         *
         * @param i0
         * @param i1
         * @param j
         * @param digits the number of digits
         * @param trailing
         * @return the decimal format
         */
        protected DecimalFormat format(int i0, int i1, int j, int digits,
                boolean trailing) {
            FlexibleDecimalFormat df = new FlexibleDecimalFormat(digits, trailing);
            df.grouping(true);
            for (int i = i0; i <= i1; i++) {
                df.update(A[i][j]);
            }
            return df;
        }

        /**
         * Decimal format for converting a matrix into a string
         *
         * @param i0
         * @param i1
         * @param j0
         * @param j1
         * @param trailing
         * @param digits the number of digits
         * @return the decimal format
         */
        protected DecimalFormat[] format(int i0, int i1, int j0, int j1,
                int digits, boolean trailing) {
            DecimalFormat[] f = new DecimalFormat[j1 - j0 + 1];
            for (int j = j0; j <= j1; j++) {
                f[j] = format(i0, i1, j, digits, trailing);
            }
            return f;
        }

        public double sum2(int j, int i0, int i1, boolean col) {
            double s2 = 0;
            if (col) {   // column 
                for (int i = i0; i <= i1; i++) {
                    s2 += A[i][j] * A[i][j];
                }
            } else {
                for (int i = i0; i <= i1; i++) {
                    s2 += A[j][i] * A[j][i];
                }
            }
            return s2;
        }

        /**
         * Squared sum of columns or rows of a matrix
         *
         * @param col if true, sum over columns; otherwise, over rows
         * @return the squared sum
         */
        public double[] sum2(boolean col) {
            int l = col ? n : m;
            int p = col ? m : n;
            double[] s2 = new double[l];
            for (int i = 0; i < l; i++) {
                s2[i] = sum2(i, 0, p - 1, col);
            }
            return s2;
        }

        /**
         * Constructs single Householder transformation for a column
         *
         * @param j the index of the column
         * @param k the index of the row
         * @return d and q
         */
        public double[] h1(int j, int k) {
            double dq[] = new double[2];
            double s2 = sum2(j, k, m - 1, true);
            dq[0] = A[k][j] >= 0 ? -Math.sqrt(s2) : Math.sqrt(s2);
            A[k][j] -= dq[0];
            dq[1] = A[k][j] * dq[0];
            return dq;
        }

        /**
         * Performs single Householder transformation on one column of a matrix
         *
         * @param j the index of the column
         * @param k the index of the row
         * @param q q = - u'u/2; must be negative
         * @param b the matrix to be transformed
         * @param l the column of the matrix b
         */
        public void h2(int j, int k, double q, NNLSMatrix b, int l) {
            double s = 0, alpha;
            for (int i = k; i < m; i++) {
                s += A[i][j] * b.A[i][l];
            }
            alpha = s / q;
            for (int i = k; i < m; i++) {
                b.A[i][l] += alpha * A[i][j];
            }
        }

        /**
         * Constructs the Givens rotation
         *
         * @param a
         * @param b
         * @return a double array that stores the cosine and sine values
         */
        public double[] g1(double a, double b) {
            double cs[] = new double[2];
            double r = Maths.hypot(a, b);
            if (r == 0.0) {
                cs[0] = 1;
                cs[1] = 0;
            } else {
                cs[0] = a / r;
                cs[1] = b / r;
            }
            return cs;
        }

        /**
         * Performs the Givens rotation
         *
         * @param cs a array storing the cosine and sine values
         * @param i0 the index of the row of the first element
         * @param i1 the index of the row of the second element
         * @param j the index of the column
         */
        public void g2(double cs[], int i0, int i1, int j) {
            double w = cs[0] * A[i0][j] + cs[1] * A[i1][j];
            A[i1][j] = -cs[1] * A[i0][j] + cs[0] * A[i1][j];
            A[i0][j] = w;
        }

        /**
         * Forward ordering of columns in terms of response explanation. On
         * input, matrices A and b are already QR-transformed. The indices of
         * transformed columns are stored in the pivoting vector.
         *
         * @param b the PaceMatrix b
         * @param pvt the pivoting vector
         * @param k0 the first k0 columns (in pvt) of A are not to be changed
         *
         */
        public void forward(NNLSMatrix b, NNLSIntVector pvt, int k0) {
            for (int j = k0; j < Math.min(pvt.size(), m); j++) {
                steplsqr(b, pvt, j, mostExplainingColumn(b, pvt, j), true);
            }
        }

        /**
         * Returns the index of the column that has the largest (squared)
         * response, when each of columns pvt[ks:] is moved to become the ks-th
         * column. On input, A and b are both QR-transformed.
         *
         * @param b response
         * @param pvt pivoting index of A
         * @param ks columns pvt[ks:] of A are to be tested
         * @return the index of the column
         */
        public int mostExplainingColumn(NNLSMatrix b, NNLSIntVector pvt, int ks) {
            double val;
            int[] p = pvt.getArray();
            double ma = columnResponseExplanation(b, pvt, ks, ks);
            int jma = ks;
            for (int i = ks + 1; i < pvt.size(); i++) {
                val = columnResponseExplanation(b, pvt, i, ks);
                if (val > ma) {
                    ma = val;
                    jma = i;
                }
            }
            return jma;
        }

        /**
         * Backward ordering of columns in terms of response explanation. On
         * input, matrices A and b are already QR-transformed. The indices of
         * transformed columns are stored in the pivoting vector.
         *
         * A and b must have the same number of rows, being the (pseudo-)rank.
         *
         * @param b PaceMatrix b
         * @param pvt pivoting vector
         * @param ks number of QR-transformed columns; psuedo-rank of A
         * @param k0 first k0 columns in pvt[] are not to be ordered.
         */
        public void backward(NNLSMatrix b, NNLSIntVector pvt, int ks, int k0) {
            for (int j = ks; j > k0; j--) {
                steplsqr(b, pvt, j, leastExplainingColumn(b, pvt, j, k0), false);
            }
        }

        /**
         * Returns the index of the column that has the smallest (squared)
         * response, when the column is moved to become the (ks-1)-th column. On
         * input, A and b are both QR-transformed.
         *
         * @param b response
         * @param pvt pivoting index of A
         * @param ks psudo-rank of A
         * @param k0 A[][pvt[0:(k0-1)]] are excluded from the testing.
         * @return the index of the column
         */
        public int leastExplainingColumn(NNLSMatrix b, NNLSIntVector pvt, int ks,
                int k0) {
            double val;
            int[] p = pvt.getArray();
            double mi = columnResponseExplanation(b, pvt, ks - 1, ks);
            int jmi = ks - 1;
            for (int i = k0; i < ks - 1; i++) {
                val = columnResponseExplanation(b, pvt, i, ks);
                if (val <= mi) {
                    mi = val;
                    jmi = i;
                }
            }
            return jmi;
        }

        /**
         * Returns the squared ks-th response value if the j-th column becomes
         * the ks-th after orthogonal transformation. A[][pvt[ks:j]] (or
         * A[][pvt[j:ks]], if ks > j) and b[] are already QR-transformed on
         * input and will remain unchanged on output.
         *
         * More generally, it returns the inner product of the corresponding row
         * vector of the response PaceMatrix. (To be implemented.)
         *
         * @param b PaceMatrix b
         * @param pvt pivoting vector
         * @param j the column A[pvt[j]][] is to be moved
         * @param ks the target column A[pvt[ks]][]
         * @return the squared response value
         */
        public double columnResponseExplanation(NNLSMatrix b, NNLSIntVector pvt,
                int j, int ks) {
            /*  Implementation: 
             *
             *  If j == ks - 1, returns the squared ks-th response directly.
             *
             *  If j > ks -1, returns the ks-th response after
             *  Householder-transforming the j-th column and the response.
             *
             *  If j < ks - 1, returns the ks-th response after a sequence of
             *  Givens rotations starting from the j-th row. */

            int k, l;
            double[] xxx = new double[n];
            int[] p = pvt.getArray();
            double val;

            if (j == ks - 1) {
                val = b.A[j][0];
            } else if (j > ks - 1) {
                int jm = Math.min(n - 1, j);
                NNLSDoubleVector u = getColumn(ks, jm, p[j]);
                NNLSDoubleVector v = b.getColumn(ks, jm, 0);
                val = v.innerProduct(u) / u.norm2();
            } else {                 // ks > j
                for (k = j + 1; k < ks; k++) // make a copy of A[j][]
                {
                    xxx[k] = A[j][p[k]];
                }
                val = b.A[j][0];
                double[] cs;
                for (k = j + 1; k < ks; k++) {
                    cs = g1(xxx[k], A[k][p[k]]);
                    for (l = k + 1; l < ks; l++) {
                        xxx[l] = -cs[1] * xxx[l] + cs[0] * A[k][p[l]];
                    }
                    val = -cs[1] * val + cs[0] * b.A[k][0];
                }
            }
            return val * val;  // or inner product in later implementation???
        }

        /**
         * QR transformation for a least squares problem<br/> A x = b<br/>
         * implicitly both A and b are transformed. pvt.size() is the
         * psuedo-rank of A.
         *
         * @param b PaceMatrix b
         * @param pvt pivoting vector
         * @param k0 the first k0 columns of A (indexed by pvt) are pre-chosen.
         * (But subject to rank examination.)
         *
         * For example, the constant term may be reserved, in which case k0 = 1.
         *
         */
        public void lsqr(NNLSMatrix b, NNLSIntVector pvt, int k0) {
            final double TINY = 1e-15;
            int[] p = pvt.getArray();
            int ks = 0;  // psuedo-rank
            for (int j = 0; j < k0; j++) // k0 pre-chosen columns
            {
                if (sum2(p[j], ks, m - 1, true) > TINY) { // large diagonal element 
                    steplsqr(b, pvt, ks, j, true);
                    ks++;
                } else {                     // collinear column
                    pvt.shiftToEnd(j);
                    pvt.setSize(pvt.size() - 1);
                    k0--;
                    j--;
                }
            }

            // initial QR transformation
            for (int j = k0; j < Math.min(pvt.size(), m); j++) {
                if (sum2(p[j], ks, m - 1, true) > TINY) {
                    steplsqr(b, pvt, ks, j, true);
                    ks++;
                } else {                     // collinear column
                    pvt.shiftToEnd(j);
                    pvt.setSize(pvt.size() - 1);
                    j--;
                }
            }

            b.m = m = ks;           // reset number of rows
            pvt.setSize(ks);
        }

        /**
         * QR transformation for a least squares problem <br/> A x = b <br/>
         * implicitly both A and b are transformed. pvt.size() is the
         * psuedo-rank of A.
         *
         * @param b PaceMatrix b
         * @param pvt pivoting vector
         * @param k0 the first k0 columns of A (indexed by pvt) are pre-chosen.
         * (But subject to rank examination.)
         *
         * For example, the constant term may be reserved, in which case k0 = 1.
         *
         */
        public void lsqrSelection(NNLSMatrix b, NNLSIntVector pvt, int k0) {
            int numObs = m;         // number of instances
            int numXs = pvt.size();

            lsqr(b, pvt, k0);

            if (numXs > 200 || numXs > numObs) { // too many columns.  
                forward(b, pvt, k0);
            }
            backward(b, pvt, pvt.size(), k0);
        }

        /**
         * Sets all diagonal elements to be positive (or nonnegative) without
         * changing the least squares solution
         *
         * @param Y the response
         * @param pvt the pivoted column index
         */
        public void positiveDiagonal(NNLSMatrix Y, NNLSIntVector pvt) {

            int[] p = pvt.getArray();
            for (int i = 0; i < pvt.size(); i++) {
                if (A[i][p[i]] < 0.0) {
                    for (int j = i; j < pvt.size(); j++) {
                        A[i][p[j]] = -A[i][p[j]];
                    }
                    Y.A[i][0] = -Y.A[i][0];
                }
            }
        }

        /**
         * Stepwise least squares QR-decomposition of the problem A x = b
         *
         * @param b PaceMatrix b
         * @param pvt pivoting vector
         * @param ks number of transformed columns
         * @param j pvt[j], the column to adjoin or delete
         * @param adjoin to adjoin if true; otherwise, to delete
         */
        public void steplsqr(NNLSMatrix b, NNLSIntVector pvt, int ks, int j,
                boolean adjoin) {
            final int kp = pvt.size(); // number of columns under consideration
            int[] p = pvt.getArray();

            if (adjoin) {     // adjoining 
                int pj = p[j];
                pvt.swap(ks, j);
                double dq[] = h1(pj, ks);
                int pk;
                for (int k = ks + 1; k < kp; k++) {
                    pk = p[k];
                    h2(pj, ks, dq[1], this, pk);
                }
                h2(pj, ks, dq[1], b, 0); // for matrix. ???
                A[ks][pj] = dq[0];
                for (int k = ks + 1; k < m; k++) {
                    A[k][pj] = 0;
                }
            } else {          // removing 
                int pj = p[j];
                for (int i = j; i < ks - 1; i++) {
                    p[i] = p[i + 1];
                }
                p[ks - 1] = pj;
                double[] cs;
                for (int i = j; i < ks - 1; i++) {
                    cs = g1(A[i][p[i]], A[i + 1][p[i]]);
                    for (int l = i; l < kp; l++) {
                        g2(cs, i, i + 1, p[l]);
                    }
                    for (int l = 0; l < b.n; l++) {
                        b.g2(cs, i, i + 1, l);
                    }
                }
            }
        }

        /**
         * Solves upper-triangular equation <br/> R x = b <br/> On output, the
         * solution is stored in b
         *
         * @param b the response
         * @param pvt the pivoting vector
         * @param kp the number of the first columns involved
         */
        public void rsolve(NNLSMatrix b, NNLSIntVector pvt, int kp) {
            if (kp == 0) {
                b.m = 0;
            }
            int i, j, k;
            int[] p = pvt.getArray();
            double s;
            double[][] ba = b.getArray();
            for (k = 0; k < b.n; k++) {
                ba[kp - 1][k] /= A[kp - 1][p[kp - 1]];
                for (i = kp - 2; i >= 0; i--) {
                    s = 0;
                    for (j = i + 1; j < kp; j++) {
                        s += A[i][p[j]] * ba[j][k];
                    }
                    ba[i][k] -= s;
                    ba[i][k] /= A[i][p[i]];
                }
            }
            b.m = kp;
        }

        /**
         * Returns a new matrix which binds two matrices together with rows.
         *
         * @param b the second matrix
         * @return the combined matrix
         */
        public NNLSMatrix rbind(NNLSMatrix b) {
            if (n != b.n) {
                throw new IllegalArgumentException("unequal numbers of rows.");
            }
            NNLSMatrix c = new NNLSMatrix(m + b.m, n);
            c.setMatrix(0, m - 1, 0, n - 1, this);
            c.setMatrix(m, m + b.m - 1, 0, n - 1, b);
            return c;
        }

        /**
         * Returns a new matrix which binds two matrices with columns.
         *
         * @param b the second matrix
         * @return the combined matrix
         */
        public NNLSMatrix cbind(NNLSMatrix b) {
            if (m != b.m) {
                throw new IllegalArgumentException("unequal numbers of rows: "
                        + m + " and " + b.m);
            }
            NNLSMatrix c = new NNLSMatrix(m, n + b.n);
            c.setMatrix(0, m - 1, 0, n - 1, this);
            c.setMatrix(0, m - 1, n, n + b.n - 1, b);
            return c;
        }

        /**
         * Solves the nonnegative linear squares problem. That is, <p> <center>
         * min || A x - b||, subject to x >= 0. </center> <p>
         *
         * For algorithm, refer to P161, Chapter 23 of C. L. Lawson and R. J.
         * Hanson (1974). "Solving Least Squares Problems". Prentice-Hall.
         *
         * @param b the response
         * @param pvt vector storing pivoting column indices
         * @return solution
         */
        public NNLSDoubleVector nnls(NNLSMatrix b, NNLSIntVector pvt) {
            int j, t, counter = 0, jm = -1, n = pvt.size();
            double ma, max, alpha, wj;
            int[] p = pvt.getArray();
            NNLSDoubleVector x = new NNLSDoubleVector(n);
            double[] xA = x.getArray();
            NNLSMatrix z = new NNLSMatrix(n, 1);
            NNLSMatrix bt;

            // step 1 
            int kp = 0; // #variables in the positive set P
            while (true) {         // step 2 
                if (++counter > 3 * n) // should never happen
                {
                    throw new RuntimeException("Does not converge");
                }
                t = -1;
                max = 0.0;
                bt = new NNLSMatrix(b.transpose());
                for (j = kp; j <= n - 1; j++) {   // W = A' (b - A x) 
                    wj = bt.times(0, kp, m - 1, this, p[j]);
                    if (wj > max) {        // step 4
                        max = wj;
                        t = j;
                    }
                }

                // step 3 
                if (t == -1) {
                    break; // optimum achieved 
                }
                // step 5 
                pvt.swap(kp, t);       // move variable from set Z to set P
                kp++;
                xA[kp - 1] = 0;
                steplsqr(b, pvt, kp - 1, kp - 1, true);
                // step 6
                ma = 0;
                while (ma < 1.5) {
                    for (j = 0; j <= kp - 1; j++) {
                        z.A[j][0] = b.A[j][0];
                    }
                    rsolve(z, pvt, kp);
                    ma = 2;
                    jm = -1;
                    for (j = 0; j <= kp - 1; j++) {  // step 7, 8 and 9
                        if (z.A[j][0] <= 0.0) { // alpha always between 0 and 1
                            alpha = xA[j] / (xA[j] - z.A[j][0]);
                            if (alpha < ma) {
                                ma = alpha;
                                jm = j;
                            }
                        }
                    }
                    if (ma > 1.5) {
                        for (j = 0; j <= kp - 1; j++) {
                            xA[j] = z.A[j][0];  // step 7 
                        }
                    } else {
                        for (j = kp - 1; j >= 0; j--) { // step 10
                            // Modified to avoid round-off error (which seemingly 
                            // can cause infinite loop).
                            if (j == jm) { // step 11 
                                xA[j] = 0.0;
                                steplsqr(b, pvt, kp, j, false);
                                kp--;  // move variable from set P to set Z
                            } else {
                                xA[j] += ma * (z.A[j][0] - xA[j]);
                            }
                        }
                    }
                }
            }
            x.setSize(kp);
            pvt.setSize(kp);
            return x;
        }

        /**
         * Solves the nonnegative least squares problem with equality
         * constraint. That is, <p> <center> min ||A x - b||, subject to x >= 0
         * and c x = d. </center> <p>
         *
         * @param b the response
         * @param c coeficients of equality constraints
         * @param d constants of equality constraints
         * @param pvt vector storing pivoting column indices
         * @return the solution
         */
        public NNLSDoubleVector nnlse(NNLSMatrix b, NNLSMatrix c, NNLSMatrix d,
                NNLSIntVector pvt) {
            double eps = 1e-10 * Math.max(c.maxAbs(), d.maxAbs())
                    / Math.max(maxAbs(), b.maxAbs());

            NNLSMatrix e = c.rbind(new NNLSMatrix(times(eps)));
            NNLSMatrix f = d.rbind(new NNLSMatrix(b.times(eps)));

            return e.nnls(f, pvt);
        }

        /**
         * Solves the nonnegative least squares problem with equality
         * constraint. That is, <p> <center> min ||A x - b||, subject to x >= 0
         * and || x || = 1. </center> <p>
         *
         * @param b the response
         * @param pvt vector storing pivoting column indices
         * @return the solution
         */
        public NNLSDoubleVector nnlse1(NNLSMatrix b, NNLSIntVector pvt) {
            NNLSMatrix c = new NNLSMatrix(1, n, 1);
            NNLSMatrix d = new NNLSMatrix(1, b.n, 1);

            return nnlse(b, c, d, pvt);
        }

        /**
         * Returns the revision string.
         *
         * @return	the revision
         */
        public String getRevision() {
            return RevisionUtils.extract("$Revision: 1.6 $");
        }
    }
}
